######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Undo/redo commands for the Tool project item."""
import copy
from spine_items.commands import SpineToolboxCommand


class UpdateToolExecuteInWorkCommand(SpineToolboxCommand):
    """Command to update Tool execute_in_work setting."""

    def __init__(self, tool_name, execute_in_work, project):
        """
        Args:
            tool_name (str): Tool's name
            execute_in_work (bool): True or False
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._tool_name = tool_name
        self._execute_in_work = execute_in_work
        self._project = project
        self.setText(f"change execute in work setting of {tool_name}")

    def redo(self):
        tool = self._project.get_item(self._tool_name)
        tool.do_update_execution_mode(self._execute_in_work)

    def undo(self):
        tool = self._project.get_item(self._tool_name)
        tool.do_update_execution_mode(not self._execute_in_work)


class UpdateToolOptionsCommand(SpineToolboxCommand):
    """Command to update Tool options."""

    def __init__(self, tool_name, changed_options, current_options, project):
        """
        Args:
            tool_name (str): Tool's name
            changed_options (dict): The options that change
            current_options (dict): Current options
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._tool_name = tool_name
        self._old_options = copy.deepcopy(current_options)
        self._new_options = copy.deepcopy(current_options)
        self._new_options.update(changed_options)
        self._project = project
        self.setText(f"change options of {tool_name}")

    def redo(self):
        tool = self._project.get_item(self._tool_name)
        tool.do_set_options(copy.deepcopy(self._new_options))
        self.setObsolete(self._old_options == self._new_options)

    def undo(self):
        tool = self._project.get_item(self._tool_name)
        tool.do_set_options(copy.deepcopy(self._old_options))


class UpdateKillCompletedProcesses(SpineToolboxCommand):
    """Command to update Tool kill_completed_processes flag."""

    def __init__(self, tool_name, kill_completed_processes, project):
        """
        Args:
            tool_name (str): Tool's name
            kill_completed_processes (bool): new flag value
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._tool_name = tool_name
        self._kill_completed_processes = kill_completed_processes
        self._project = project
        self.setText(f"change kill consoles setting of {tool_name}")

    def redo(self):
        tool = self._project.get_item(self._tool_name)
        tool.do_update_kill_completed_processes(self._kill_completed_processes)

    def undo(self):
        tool = self._project.get_item(self._tool_name)
        tool.do_update_kill_completed_processes(not self._kill_completed_processes)


class UpdateLogProcessOutput(SpineToolboxCommand):
    """Command to update Tool log_process_output flag."""

    def __init__(self, tool_name, log_process_output, project):
        """
        Args:
            tool_name (str): Tool's name
            log_process_output (bool): new flag value
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._tool_name = tool_name
        self._log_process_output = log_process_output
        self._project = project
        self.setText(f"change log process output setting of {tool_name}")

    def redo(self):
        tool = self._project.get_item(self._tool_name)
        tool.do_update_log_process_output(self._log_process_output)

    def undo(self):
        tool = self._project.get_item(self._tool_name)
        tool.do_update_log_process_output(not self._log_process_output)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains ConnectionManager class."""
from PySide6.QtCore import QObject, Qt, QThread, Signal, Slot
from spinetoolbox.helpers import busy_effect


class ConnectionManager(QObject):
    """Class to manage data connections in another thread."""

    tables_requested = Signal()
    data_requested = Signal(str, dict, int, int)
    default_mapping_requested = Signal()

    connection_failed = Signal(str)
    """Signal with error message if connection fails  """

    connection_ready = Signal()
    """Signal that a connection to the datasource is ready  """

    connection_closed = Signal()
    """Signal that connection is being closed  """

    error = Signal(str)
    """error while reading data or connection to data source  """

    data_ready = Signal(list, list)
    """data from source is ready, should send list of data and headers  """

    tables_ready = Signal(list)
    """tables from source is ready, should send a list of str of available tables  """

    default_mapping_ready = Signal(dict)
    """default mapping ready from data source  """

    current_table_changed = Signal()
    """Emitted when the current table has changed."""

    def __init__(self, connection, connection_settings, parent):
        """
        Args:
            connection (Type): A class derived from `SourceConnection`, e.g. `CSVConnector`
            connection_settings (dict): connection specific settings
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._thread = None
        self._worker = None
        self._current_table = None
        self._table_options = {}
        self._table_types = {}
        self._default_table_column_type = {}
        self._table_row_types = {}
        self._connection = connection
        self._connection_settings = connection_settings
        self._is_connected = False

    @property
    def connection(self):
        return self._connection

    @property
    def current_table(self):
        return self._current_table

    @property
    def is_connected(self):
        return self._is_connected

    @property
    def table_options(self):
        return self._table_options

    @property
    def table_types(self):
        return self._table_types

    @property
    def table_default_column_type(self):
        return self._default_table_column_type

    @property
    def table_row_types(self):
        return self._table_row_types

    @property
    def source_type(self):
        return self._connection.__name__

    def set_table(self, table):
        """Sets the current table of the data source.

        Args:
            table (str): table name
        """
        self._current_table = table
        self.current_table_changed.emit()

    def request_tables(self):
        """Gets tables tables from source, emits tables_requested."""
        if self.is_connected:
            self.tables_requested.emit()

    def request_data(self, table=None, max_rows=-1, start=0):
        """Requests data from a table. Starts a process that emits data_ready with the fetched data.

        Args:
            table (str): which table to get data from (default: {None})
            max_rows (int): how many rows to read (default: {-1})
        """
        if not self.is_connected:
            return
        if table is None:
            table = self._current_table
        options = self._table_options.get(self._current_table, {})
        self.data_requested.emit(table, options, max_rows, start)

    def request_default_mapping(self):
        """Request default mapping from worker."""
        if self.is_connected:
            self.default_mapping_requested.emit()

    def init_connection(self, source, **source_extras):
        """Creates a Worker and a new thread to read source data.
        If there is an existing thread close that one.

        Args:
            source (str): source file name or URL
            **source_extras: source specific additional connection settings
        """
        self.close_connection()
        self._worker = ConnectionWorker(self._connection, self._connection_settings, source, source_extras)
        self._thread = QThread()
        self._worker.moveToThread(self._thread)
        self._worker.connectionReady.connect(self._handle_connection_ready)
        self._worker.tablesReady.connect(self._handle_tables_ready)
        self._worker.dataReady.connect(self.data_ready.emit)
        self._worker.defaultMappingReady.connect(self.default_mapping_ready.emit)
        self._worker.error.connect(self.error.emit)
        self._worker.connectionFailed.connect(self.connection_failed.emit)
        self.tables_requested.connect(self._worker.tables)
        self.data_requested.connect(self._worker.data)
        self.default_mapping_requested.connect(self._worker.default_mapping)
        self.connection_closed.connect(self._worker.close_connection, type=Qt.ConnectionType.BlockingQueuedConnection)
        self._thread.started.connect(self._worker.init_connection)
        self._thread.start()

    @Slot()
    def _handle_connection_ready(self):
        self._is_connected = True
        self.connection_ready.emit()

    @Slot(dict)
    def _handle_tables_ready(self, table_properties):
        for key, properties in table_properties.items():
            if key not in self._table_options:
                self._table_options[key] = properties.options
            if key not in self._table_types:
                self._table_types[key] = {}
            if key not in self._table_row_types:
                self._table_row_types[key] = {}
        self.tables_ready.emit(list(table_properties))
        # update options if a sheet is selected
        if self._current_table in self._table_options:
            self.current_table_changed.emit()

    @Slot(dict)
    def update_options(self, options):
        if not self._current_table:
            return
        self._table_options.setdefault(self._current_table, {}).update(options)

    def get_current_options(self):
        if not self._current_table:
            return {}
        return self._table_options.get(self._current_table, {})

    def get_current_option_value(self, option_key):
        """Returns the value for option_key for the current table or the default value."""
        current_options = self._table_options.get(self._current_table, {})
        option_value = current_options.get(option_key)
        if option_value is None:
            option_specification = self._connection.OPTIONS.get(option_key)
            if option_specification is not None:
                return option_specification["default"]
            option_specification = self._connection.BASE_OPTIONS[option_key]
            return option_specification["default"]
        return option_value

    def set_table_options(self, options):
        """Sets connection manager options for current connector

        Args:
            options (dict): settings for the tables
        """
        self._table_options.update(options)
        if self._current_table in self._table_options:
            self.current_table_changed.emit()

    def set_table_types(self, types):
        """Sets connection manager types for current connector

        Args:
            types (dict): dict with types settings, column (int) as key, type as value
        """
        self._table_types.update(types)

    def update_table_default_column_type(self, column_type):
        """Updates default column type.

        Args:
            column_type (dict): mapping from table name to column type name
        """
        self._default_table_column_type.update(column_type)

    def clear_table_default_column_type(self, table_name):
        """Clears default column type.

        Args:
            table_name (str): table name
        """
        self._default_table_column_type.pop(table_name, None)

    def set_table_row_types(self, types):
        """Sets connection manager types for current connector

        Arguments:
            types {dict} -- Dict with types settings, row (int) as key, type as value
        """
        self._table_row_types.update(types)

    def close_connection(self):
        """Closes and deletes thread and worker"""
        self._is_connected = False
        self.connection_closed.emit()
        if self._worker:
            self.connection_closed.disconnect(self._worker.close_connection)
            self._worker.deleteLater()
            self._worker = None
        if self._thread:
            self._thread.quit()
            self._thread.wait()
            self._thread.deleteLater()


class ConnectionWorker(QObject):
    """A class for delegating SourceConnection operations to another QThread."""

    connectionFailed = Signal(str)
    """Signal with error message if connection fails"""
    error = Signal(str)
    """Signal with error message if something errors"""
    connectionReady = Signal()
    """Signal that connection is ready to be read"""
    tablesReady = Signal(dict)
    """Signal when tables from source is ready, a list of table names or dict mapping table name to table options."""
    dataReady = Signal(list, list)
    """Signal when data from a specific table in source is ready, list of data and list of headers"""
    defaultMappingReady = Signal(dict)
    """Signal when default mapping is ready"""

    def __init__(self, connection, connection_settings, source, source_extras, parent=None):
        """
        Args:
            connection (class): A class derived from `SourceConnection` for connecting to the source file
            connection_settings (dict): settings passed to the connection constructor
            source (str): source file path or URL
            source_extras (dict): source specific additional connection settings
            parent (QObject, optional): parent object
        """
        super().__init__(parent)
        self._connection = connection(connection_settings)
        self._source = source
        self._source_extras = source_extras

    @Slot()
    def init_connection(self):
        """Connects to data source."""
        if self._source:
            try:
                self._connection.connect_to_source(self._source, **self._source_extras)
                self.connectionReady.emit()
            except Exception as error:
                self.connectionFailed.emit(f"Could not connect to source: {error}")
        else:
            self.connectionFailed.emit("Connection has no source")

    @Slot()
    def tables(self):
        try:
            table_properties = self._connection.get_tables_and_properties()
            self.tablesReady.emit(table_properties)
        except Exception as error:
            self.error.emit(f"Could not get tables from source: {error}")

    @busy_effect
    @Slot(str, dict, int, int)
    def data(self, table, options, max_rows, start):
        if not table:
            # FIXME: The 'Select all' option in the Source tables list thinks it's a table too and requests data
            return
        try:
            data, header = self._connection.get_data(table, options, max_rows, start)
            self.dataReady.emit(data, header)
        except Exception as error:
            self.error.emit(f"Could not get data from source: {error}")

    @Slot()
    def close_connection(self):
        try:
            self._connection.disconnect()
        except Exception as error:
            self.error.emit(f"Could not disconnect from source: {error}")

    @Slot()
    def default_mapping(self):
        try:
            mapping = self._connection.create_default_mapping()
            self.defaultMappingReady.emit(mapping)
        except Exception as error:
            self.error.emit(f"Could not default mapping from source: {error}")



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains the :class:`Exporter` project item."""
from dataclasses import dataclass
from itertools import combinations, zip_longest
from operator import itemgetter
from pathlib import Path
from PySide6.QtCore import Qt, Slot
from spine_engine.utils.helpers import ExecutionDirection
from spine_engine.utils.serialization import deserialize_path
from spinedb_api import clear_filter_configs
from spinetoolbox.helpers import SealCommand
from spinetoolbox.project_item.project_item import ProjectItem
from ..commands import UpdateCancelOnErrorCommand
from .commands import CommandId, UpdateOutLabel, UpdateOutputTimeStampsFlag, UpdateOutUrl
from .executable_item import ExecutableItem
from .export_manifest import exported_files_as_resources, is_manifest_file
from .item_info import ItemInfo
from .mvcmodels.full_url_list_model import FullUrlListModel
from .output_channel import OutputChannel
from .specification import OutputFormat
from .utils import output_database_resources
from .widgets.export_list_item import ExportListItem


@dataclass
class _Notifications:
    """Holds flags for different exporter error conditions."""

    duplicate_out_label: bool = False
    missing_out_label: bool = False
    unrecognized_extension: bool = False
    missing_specification: bool = False


class Exporter(ProjectItem):
    """Exporter project item."""

    def __init__(
        self,
        name,
        description,
        x,
        y,
        toolbox,
        project,
        specification_name=None,
        output_channels=None,
        output_time_stamps=False,
        cancel_on_error=True,
    ):
        """
        Args:
            name (str): item's name
            description (str): item's description
            x (float): Initial X coordinate of item icon
            y (float): Initial Y coordinate of item icon
            toolbox (QWidget): parent window
            project (SpineToolboxProject): the project this item belongs to
            specification_name (str, optional): exporter specification
            output_channels (list of OutputChannel, optional): input and output labels
            output_time_stamps (bool): True to include time stamps to output directory names
            cancel_on_error (bool): True to fail execution in case of non-fatal errors
        """
        super().__init__(name, description, x, y, project)
        self._toolbox = toolbox
        self._append_output_time_stamps = output_time_stamps
        self._cancel_on_error = cancel_on_error
        self._output_filenames = {}
        self._export_list_items = {}
        self._full_url_model = FullUrlListModel()
        self._exported_files = None
        self._notifications = _Notifications()
        self._output_channels = []
        if output_channels is None:
            output_channels = []
        self._inactive_output_channels = output_channels
        self._specification_name = specification_name
        self._specification = self._project.get_specification(specification_name)
        if specification_name and not self._specification:
            self._toolbox.msg_error.emit(
                f"Exporter <b>{self.name}</b> should have a specification <b>{specification_name}</b> but it was not found"
            )
        self.do_set_specification(self._specification)

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    def has_out_url(self):
        """Returns whether any output channel has a URL set.

        Returns:
            bool: True if any channel has output URL, False otherwise
        """
        return any(c.out_url is not None for c in self._output_channels)

    def handle_execution_successful(self, execution_direction):
        """See base class."""
        if execution_direction != ExecutionDirection.FORWARD:
            return
        self._resources_to_successors_changed()

    def full_url_model(self):
        """
        Returns the full URL model held by the exporter.

        Returns:
            FullUrlListModel: full URL model
        """
        return self._full_url_model

    @Slot(int)
    def _cancel_on_error_option_changed(self, checkbox_state):
        """Handles changes to the Cancel export on error option."""
        cancel = checkbox_state == Qt.CheckState.Checked.value
        if self._cancel_on_error == cancel:
            return
        self._toolbox.undo_stack.push(UpdateCancelOnErrorCommand(self.name, cancel, self._project))

    def set_cancel_on_error(self, cancel):
        """Sets the Cancel export on error option."""
        self._cancel_on_error = cancel
        if not self._active:
            return
        # This does not trigger the stateChanged signal.
        self._properties_ui.cancel_on_error_check_box.setCheckState(
            Qt.CheckState.Checked if cancel else Qt.CheckState.Unchecked
        )

    def _update_properties_tab(self):
        """Updates the labels list in the properties tab."""
        self._set_properties_message()
        outputs_layout = self._properties_ui.outputs_list_layout
        while not outputs_layout.isEmpty():
            widget_to_remove = outputs_layout.takeAt(0)
            widget_to_remove.widget().deleteLater()
        self._export_list_items.clear()
        for channel in self._output_channels:
            item = self._export_list_items[channel.in_label] = ExportListItem(
                channel.in_label, channel.out_label, channel.out_url, self._toolbox.qsettings(), self._toolbox
            )
            outputs_layout.addWidget(item)
            output_format = self._specification.output_format if self._specification is not None else None
            item.set_out_url_enabled(output_format is None or output_format == OutputFormat.SQL)
            item.out_label_changed.connect(self._update_out_label)
            item.out_label_editing_finished.connect(self._seal_out_label_update)
            item.out_url_changed.connect(self._update_out_url)
        self._properties_ui.output_time_stamps_check_box.setCheckState(
            Qt.CheckState.Checked if self._append_output_time_stamps else Qt.CheckState.Unchecked
        )
        self._properties_ui.cancel_on_error_check_box.setCheckState(
            Qt.CheckState.Checked if self._cancel_on_error else Qt.CheckState.Unchecked
        )

    def _set_properties_message(self):
        if self._specification is None:
            message = ""
        else:
            if self._specification.is_exporting_multiple_files():
                message = (
                    f"Currently exporting multiple files in {self._specification.output_format.value} format."
                    " The file names are given by the specification."
                )
            else:
                message = (
                    f"Currently exporting in {self._specification.output_format.value} format."
                    " The Output labels below are treated as file names."
                )
        self._properties_ui.message_label.setText(message)

    def upstream_resources_updated(self, resources):
        """See base class."""
        database_resources = [r for r in resources if r.type_ == "database"]
        resources_by_labels = {r.label: r for r in database_resources}  # legacy: when in_label was url
        in_labels = set(r.label for r in database_resources)
        inactive_channels = {c.in_label: c for c in self._output_channels}
        old_output_channels = self._output_channels
        self._output_channels = []
        inactive_channels.update({c.in_label: c for c in self._inactive_output_channels})
        for in_label in in_labels:
            inactive = inactive_channels.pop(in_label, None)
            if inactive is not None:
                self._output_channels.append(inactive)
            else:
                url = clear_filter_configs(resources_by_labels[in_label].url)
                if url in inactive_channels:
                    # legacy: when in_label was url
                    # we actually have out_label already
                    inactive = inactive_channels.pop(url)
                    self._output_channels.append(OutputChannel(in_label, self.name, inactive.out_label))
                else:
                    self._output_channels.append(OutputChannel(in_label, self.name))
        self._inactive_output_channels = list(inactive_channels.values())
        self._full_url_model.set_urls(set(r.url for r in database_resources))
        if self._output_channels != old_output_channels:
            self._resources_to_successors_changed()
        if self._active:
            self._update_properties_tab()
        self._check_notifications()

    def replace_resources_from_upstream(self, old, new):
        """See base class."""
        for old_resource, new_resource in zip(old, new):
            if old_resource.type_ != "database":
                continue
            for channel in self._output_channels:
                if channel.in_label == old_resource.label:
                    channel.in_label = new_resource.label
                    break
            else:
                raise RuntimeError(f"Logic error: cannot find channel for {old_resource.label}")
            self._full_url_model.update_url(old_resource.url, new_resource.url)

    def _check_notifications(self):
        """See base class."""
        self._check_missing_out_labels()
        self._check_duplicate_out_labels()
        self._check_unrecognized_extension()
        self._check_missing_specification()
        self._report_notifications()

    def _report_notifications(self):
        """Updates the exclamation icon and notifications labels."""
        if self._icon is None:
            return
        self.clear_notifications()
        if self._notifications.duplicate_out_label:
            self.add_notification("Duplicate output labels.")
        if self._notifications.missing_out_label:
            self.add_notification("Output label(s) missing.")
        if self._notifications.missing_specification:
            self.add_notification("Export specification missing.")
        if self._notifications.unrecognized_extension:
            self.add_notification("File extensions don't match the output format.")

    def _check_missing_out_labels(self):
        """Checks the status of out labels"""
        self._notifications.missing_out_label = not all(bool(c.out_label) for c in self._output_channels)

    def _check_duplicate_out_labels(self):
        """Checks for duplicate output file names."""
        self._notifications.duplicate_out_label = any(
            c1.out_label == c2.out_label
            for c1, c2 in combinations(self._output_channels, 2)
            if c1.out_label and c2.out_label
        )

    def _check_unrecognized_extension(self):
        """Checks that file extensions match with selected output format."""
        if self._notifications.missing_specification or self._notifications.missing_out_label:
            self._notifications.unrecognized_extension = False
            return
        output_formats = set()
        for channel in self._output_channels:
            _, separator, extension = channel.out_label.rpartition(".")
            if not separator:
                continue
            output_format = OutputFormat.output_format_from_extension(extension)
            if output_format is None:
                self._notifications.unrecognized_extension = True
                return
            output_formats.add(output_format)
        if len(output_formats) == 0:
            self._notifications.unrecognized_extension = False
            return
        if len(output_formats) > 1:
            self._notifications.unrecognized_extension = True
            return
        output_format = next(iter(output_formats))
        if self._specification is not None:
            self._notifications.unrecognized_extension = output_format != self._specification.output_format

    @Slot(str, str)
    def _update_out_label(self, out_label, in_label):
        """Pushes UpdateOutLabel command to undo stack.

        Args:
            out_label (str): new out label
            in_label (str): associated in label
        """
        previous = next(c for c in self._output_channels if c.in_label == in_label)
        self._toolbox.undo_stack.push(UpdateOutLabel(self.name, out_label, in_label, previous.out_label, self._project))

    @Slot(str)
    def _seal_out_label_update(self, in_label):
        """Pushes a sealing command to undo stack.

        Args:
            in_label (str): associated in label
        """
        self._toolbox.undo_stack.push(SealCommand(CommandId.CHANGE_OUT_LABEL.value))

    @Slot(str, dict)
    def _update_out_url(self, in_label, url_dict):
        """Pushes a command to change the output URL to undo stack.

        Args:
            in_label (str): input label
            url_dict (dict): URL
        """
        for channel in self._output_channels:
            if channel.in_label == in_label:
                if channel.out_url != url_dict:
                    self._toolbox.undo_stack.push(
                        UpdateOutUrl(self.name, in_label, url_dict, channel.out_url, self._project)
                    )
                break
        else:
            raise RuntimeError(f"Logic error: cannot find channel for input label {in_label}")

    def set_out_url(self, in_label, url_dict):
        """Sets output URL for an output channel.

        Args:
            in_label (str): input label
            url_dict (dict): URL
        """
        for channel in self._output_channels:
            if channel.in_label == in_label:
                channel.out_url = url_dict
                if self._active:
                    export_list_item = self._export_list_items[in_label]
                    export_list_item.set_out_url(url_dict)
                break
        else:
            raise RuntimeError(f"Logic error: cannot find channel for input label {in_label}")

    def set_out_label(self, out_label, in_label):
        """Updates the output label.

        Args:
            out_label (str): output label
            in_label (str): corresponding input label
        """
        if self._active:
            export_list_item = self._export_list_items[in_label]
            export_list_item.out_label_edit.setText(out_label)
        channel = next(c for c in self._output_channels if c.in_label == in_label)
        old_out_label = channel.out_label
        old_resources = self.resources_for_direct_successors() if old_out_label else []
        channel.out_label = out_label
        self._check_missing_out_labels()
        self._check_duplicate_out_labels()
        self._check_unrecognized_extension()
        self._report_notifications()
        if self._exported_files is not None:
            exported_files = self._exported_files.pop(old_out_label, None)
            if exported_files is not None and out_label:
                self._exported_files[out_label] = exported_files
        if not old_out_label or not out_label:
            self._resources_to_successors_changed()
        else:
            new_resources = self.resources_for_direct_successors()
            olds = [resource for resource in old_resources if resource.label == old_out_label]
            news = [resource for resource in new_resources if resource.label == out_label]
            self._resources_to_successors_replaced(olds, news)

    def get_out_labels(self):
        """Returns output labels.

        Returns:
            set: output labels
        """
        return {channel.out_label for channel in self._output_channels}

    def item_dict(self):
        """See base class."""
        serialized = super().item_dict()
        serialized["output_time_stamps"] = self._append_output_time_stamps
        serialized["cancel_on_error"] = self._cancel_on_error
        serialized["output_labels"] = sorted(
            [c.to_dict(self._project.project_dir) for c in self._output_channels], key=itemgetter("in_label")
        )
        if self._specification is not None and self._specification.output_format == OutputFormat.SQL:
            for channel in self._output_channels:
                if channel.out_url is not None:
                    serialized.setdefault("db_credentials", {})[channel.in_label] = {
                        "username": channel.out_url.get("username"),
                        "password": channel.out_url.get("password"),
                    }
        serialized["specification"] = self._specification_name
        return serialized

    @staticmethod
    def item_dict_local_entries():
        """See base class."""
        return [("db_credentials",)]

    def resources_for_direct_successors(self):
        """See base class."""
        output_format = self._specification.output_format if self._specification is not None else None
        resources, self._exported_files = exported_files_as_resources(
            self.name, self._exported_files, self.data_dir, self._output_channels, output_format
        )
        if output_format == OutputFormat.SQL:
            resources += output_database_resources(self.name, self._output_channels)
        return resources

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        """See base class."""
        description, x, y = ProjectItem.parse_item_dict(item_dict)
        output_channels = [
            OutputChannel.from_dict(d, name, project.project_dir) for d in item_dict.get("output_labels", [])
        ]
        for db_dict in item_dict.get("databases", []):
            # Legacy item dict.
            out_label = db_dict["output_file_name"]
            url = clear_filter_configs(deserialize_path(db_dict["database_url"], project.project_dir))
            output_channels.append(OutputChannel(url, name, out_label))
        for in_label, credentials in item_dict.get("db_credentials", {}).items():
            for channel in output_channels:
                if channel.in_label == in_label:
                    channel.out_url.update(credentials)
        output_time_stamps = item_dict.get("output_time_stamps", False)
        cancel_on_error = item_dict.get("cancel_on_error", True)
        specification_name = item_dict.get("specification", "")
        specification = project.get_specification(specification_name)
        if specification_name and not specification:
            toolbox.msg_error.emit(f"Exporter <b>{name}</b> specification <b>{specification_name}</b> not found.")
        return Exporter(
            name,
            description,
            x,
            y,
            toolbox,
            project,
            specification_name,
            output_channels,
            output_time_stamps,
            cancel_on_error,
        )

    def rename(self, new_name, rename_data_dir_message):
        """See base class."""
        if not super().rename(new_name, rename_data_dir_message):
            return False
        for path in Path(self.data_dir).iterdir():
            if is_manifest_file(path):
                path.unlink()
        if self._exported_files is not None:
            data_dir_parts = Path(self.data_dir).parts
            for label, file_list in self._exported_files.items():
                new_file_list = []
                for file_path in file_list:
                    new_file_path = Path()
                    for old_part, new_part in zip_longest(Path(file_path).parts, data_dir_parts):
                        if new_part is None:
                            new_file_path = new_file_path / old_part
                        else:
                            new_file_path = new_file_path / new_part
                    new_file_list.append(str(new_file_path))
                self._exported_files[label] = new_file_list
        return True

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.output_time_stamps_check_box.stateChanged] = self._change_output_time_stamps_flag
        s[self._properties_ui.cancel_on_error_check_box.stateChanged] = self._cancel_on_error_option_changed
        s[self._properties_ui.specification_button.clicked] = self.show_specification_window
        s[self._properties_ui.specification_combo_box.textActivated] = self._change_specification
        return s

    @Slot(int)
    def _change_output_time_stamps_flag(self, checkbox_state):
        """
        Pushes a command that changes the output time stamps flag value.

        Args:
            checkbox_state (int): setting's checkbox state on properties tab
        """
        flag = checkbox_state == Qt.CheckState.Checked.value
        if flag == self._append_output_time_stamps:
            return
        self._toolbox.undo_stack.push(UpdateOutputTimeStampsFlag(self.name, flag, self._project))

    def set_output_time_stamps_flag(self, flag):
        """
        Sets the output time stamps flag.

        Args:
            flag (bool): flag value
        """
        self._append_output_time_stamps = flag
        if self._active:
            self._properties_ui.output_time_stamps_check_box.setChecked(flag)

    def _check_missing_specification(self):
        """Checks specification's status."""
        self._notifications.missing_specification = not self._specification_name

    def do_set_specification(self, specification):
        """see base class"""
        if not super().do_set_specification(specification):
            return False
        if specification is None:
            if self._active:
                self._properties_ui.specification_combo_box.setCurrentIndex(-1)
            return True
        self._specification_name = specification.name
        self._check_notifications()
        if self._active:
            self._properties_ui.specification_combo_box.setCurrentText(self._specification_name)
            self._set_properties_message()
            output_format = specification.output_format if specification is not None else None
            out_url_enabled = output_format is None or output_format == OutputFormat.SQL
            for item in self._export_list_items.values():
                item.set_out_url_enabled(out_url_enabled)
        return True

    @Slot(bool)
    def show_specification_window(self, _=True):
        """Opens the settings window."""
        self._toolbox.show_specification_form(self.item_type(), self.specification(), self)

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Data Store":
            self._toolbox.msg.emit(
                "Link established. You can now export the database in "
                f"<b>{source_item.name}</b> in <b>{self.name}</b>."
            )
        elif source_item.item_type() == "Data Transformer":
            self._toolbox.msg.emit(
                f"Link established. You can now export the database transformed by <b>{source_item.name}</b> "
                f"in <b>{self.name}</b>."
            )
        else:
            super().notify_destination(source_item)

    @Slot(str)
    def _change_specification(self, specification_name):
        """
        Pushes a set specification command to the undo stack.

        Args:
            specification_name (str): new specification name
        """
        specification = self._project.get_specification(specification_name)
        self.set_specification(specification)

    def restore_selections(self):
        """See base class."""
        self._update_properties_tab()
        if self._specification_name:
            self._properties_ui.specification_combo_box.setCurrentText(self._specification_name)
        else:
            self._properties_ui.specification_combo_box.setCurrentIndex(-1)

    def tear_down(self):
        super().tear_down()
        self._full_url_model.deleteLater()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Widget shown to user when a new View is created."""
from spinetoolbox.widgets.add_project_item_widget import AddProjectItemWidget
from ..item_info import ItemInfo
from ..view import View


class AddViewWidget(AddProjectItemWidget):
    """A widget to query user's preferences for a new item."""

    def __init__(self, toolbox, x, y, spec=""):
        """
        Args:
            toolbox (ToolboxUI): Parent widget
            x (int): X coordinate of new item
            y (int): Y coordinate of new item
            spec (str): item specification's name
        """
        super().__init__(toolbox, x, y, View, spec=spec)

    def call_add_item(self):
        """Creates new Item according to user's selections."""
        item = {self.name: {"type": ItemInfo.item_type(), "description": self.description, "x": self._x, "y": self._y}}
        self._toolbox.add_project_items(item)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for Exporter's :func:`do_work` function."""
from csv import reader
import os.path
import sqlite3
from tempfile import TemporaryDirectory
import unittest
from unittest.mock import MagicMock
from spine_items.exporter.do_work import do_work
from spine_items.exporter.specification import MappingSpecification, MappingType, OutputFormat, Specification
from spinedb_api import DatabaseMapping, import_object_classes, import_objects
from spinedb_api.export_mapping import entity_export
from spinedb_api.export_mapping.export_mapping import FixedValueMapping
from spinedb_api.export_mapping.group_functions import NoGroup
from spinedb_api.mapping import Position


class TestWithCsvWriter(unittest.TestCase):
    _temp_dir = None
    _url = None

    @classmethod
    def setUpClass(cls):
        cls._temp_dir = TemporaryDirectory()
        db_file = os.path.join(cls._temp_dir.name, "test_db.sqlite")
        cls._url = "sqlite:///" + db_file
        with DatabaseMapping(cls._url, create=True) as db_map:
            import_object_classes(db_map, ("oc1", "oc2"))
            import_objects(db_map, (("oc1", "o11"), ("oc1", "o12"), ("oc2", "o21"), ("oc2", "o22"), ("oc2", "o23")))
            db_map.commit_session("Add test data.")

    def test_export_database(self):
        root_mapping = entity_export(entity_class_position=0, entity_position=1)
        mapping_specification = MappingSpecification(
            MappingType.entities, True, True, NoGroup.NAME, False, root_mapping
        )
        specification = Specification("name", "description", {"mapping": mapping_specification})
        databases = {self._url: "test_export_database.csv"}
        logger = MagicMock()
        self.assertTrue(
            do_work(None, specification.to_dict(), False, False, "", self._temp_dir.name, databases, {}, "", "", logger)
        )
        out_path = os.path.join(self._temp_dir.name, "test_export_database.csv")
        self.assertTrue(os.path.exists(out_path))
        with open(out_path) as input_:
            csv_reader = reader(input_)
            table = list(csv_reader)
        expected = [["oc1", "o11"], ["oc1", "o12"], ["oc2", "o21"], ["oc2", "o22"], ["oc2", "o23"]]
        self.assertEqual(table, expected)

    def test_export_to_output_database(self):
        object_root = entity_export(entity_class_position=0, entity_position=1)
        object_root.header = "object_class"
        object_root.child.header = "object"
        root_mapping = FixedValueMapping(Position.table_name, "data_table")
        root_mapping.child = object_root
        mapping_specification = MappingSpecification(
            MappingType.entities, True, True, NoGroup.NAME, False, root_mapping
        )
        specification = Specification("name", "description", {"mapping": mapping_specification}, OutputFormat.SQL)
        databases = {self._url: "output label"}
        out_path = os.path.join(self._temp_dir.name, "out_database.sqlite")
        out_urls = {self._url: {"dialect": "sqlite", "database": out_path}}
        logger = MagicMock()
        self.assertTrue(
            do_work(
                None,
                specification.to_dict(),
                False,
                False,
                "",
                self._temp_dir.name,
                databases,
                out_urls,
                "",
                "",
                logger,
            )
        )
        self.assertTrue(os.path.exists(out_path))
        connection = sqlite3.connect(out_path)
        cursor = connection.cursor()
        expected = [("oc1", "o11"), ("oc1", "o12"), ("oc2", "o21"), ("oc2", "o22"), ("oc2", "o23")]
        self.assertEqual(cursor.execute("SELECT * FROM data_table").fetchall(), expected)
        connection.close()


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains the :class:`DataTransformer` project item."""
from json import dump
from PySide6.QtCore import Slot
from spinetoolbox.project_item.project_item import ProjectItem
from spinetoolbox.widgets.custom_menus import ItemSpecificationMenu
from .executable_item import ExecutableItem
from .filter_config_path import filter_config_path
from .item_info import ItemInfo
from .output_resources import scan_for_resources


class DataTransformer(ProjectItem):
    """Data transformer project item."""

    def __init__(self, name, description, x, y, toolbox, project, specification_name=""):
        """
        Args:
            name (str): item's name
            description (str): item's description
            x (float): Initial X coordinate of item icon
            y (float): Initial Y coordinate of item icon
            toolbox (QWidget): parent window
            project (SpineToolboxProject): the project this item belongs to
            specification_name (str, optional): transformer specification
        """
        super().__init__(name, description, x, y, project)
        self._toolbox = toolbox
        self._specification_name = specification_name
        self._specification = self._project.get_specification(specification_name)
        if specification_name and not self._specification:
            self._logger.msg_error.emit(
                f"Data Transformer <b>{self.name}</b> should have a specification <b>{specification_name}</b> but it was not found"
            )
        self._db_resources = []

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    def item_dict(self):
        """See base class."""
        serialized = super().item_dict()
        serialized["specification"] = self._specification_name
        return serialized

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        """See base class."""
        description, x, y = ProjectItem.parse_item_dict(item_dict)
        specification_name = item_dict["specification"]
        return DataTransformer(name, description, x, y, toolbox, project, specification_name)

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.specification_button.clicked] = self.show_specification_window
        s[self._properties_ui.specification_combo_box.textActivated] = self._change_specification
        return s

    def do_set_specification(self, specification):
        """see base class"""
        self._dump_filter_configs()
        old_resources = scan_for_resources(
            self, self.specification(), self._db_resources, filter_config_path(self.data_dir)
        )
        if not super().do_set_specification(specification):
            return False
        self._specification_name = specification.name if specification is not None else ""
        if self._active:
            self._update_ui()
        self._dump_filter_configs()
        new_resources = scan_for_resources(
            self, self.specification(), self._db_resources, filter_config_path(self.data_dir)
        )
        self._resources_to_successors_replaced(old_resources, new_resources)
        self._check_notifications()
        return True

    def _dump_filter_configs(self):
        """writes filter configs to disk."""
        if self._specification is not None and self._specification.settings is not None:
            path = filter_config_path(self.data_dir)
            with open(path, "w") as filter_config_file:
                dump(self._specification.settings.filter_config(), filter_config_file)

    @Slot(bool)
    def show_specification_window(self, _=True):
        """Opens the settings window."""
        self._toolbox.show_specification_form(
            self.item_type(), self.specification(), item=self, urls=[r.url for r in self._db_resources]
        )

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Data Store":
            self._logger.msg.emit(
                "Link established. You can now define transformations for Data Store "
                f"<b>{source_item.name}</b> in Data Transformer <b>{self.name}</b>."
            )
        elif source_item.item_type() == "Data Transformer":
            self._logger.msg.emit(
                "Link established. You can now define additional transformations "
                f"in Data Transformer <b>{self.name}</b>."
            )
        else:
            super().notify_destination(source_item)

    @Slot(str)
    def _change_specification(self, specification_name):
        """
        Pushes a set specification command to the undo stack.

        Args:
            specification_name (str): new specification name
        """
        specification = self._project.get_specification(specification_name)
        self.set_specification(specification)

    def upstream_resources_updated(self, resources):
        """See base class."""
        self._db_resources = [r for r in resources if r.type_ == "database"]
        self._resources_to_successors_changed()

    def replace_resources_from_upstream(self, old, new):
        """See base class."""
        old_resources = scan_for_resources(
            self, self.specification(), self._db_resources, filter_config_path(self.data_dir)
        )
        for old_resource, new_resource in zip(old, new):
            for i, resource in enumerate(self._db_resources):
                if resource == old_resource:
                    self._db_resources[i] = new_resource
                    break
        new_resources = scan_for_resources(
            self, self.specification(), self._db_resources, filter_config_path(self.data_dir)
        )
        self._resources_to_successors_replaced(old_resources, new_resources)

    def resources_for_direct_successors(self):
        """See base class."""
        return scan_for_resources(self, self.specification(), self._db_resources, filter_config_path(self.data_dir))

    def restore_selections(self):
        """See base class."""
        self._update_ui()

    def _update_ui(self):
        if self._specification_name:
            self._properties_ui.specification_combo_box.setCurrentText(self._specification_name)
            spec_model_index = self._toolbox.specification_model.specification_index(self._specification_name)
            specification_options_popup_menu = ItemSpecificationMenu(self._toolbox, spec_model_index, self)
            self._properties_ui.specification_button.setMenu(specification_options_popup_menu)
        else:
            self._properties_ui.specification_combo_box.setCurrentIndex(-1)
            self._properties_ui.specification_button.setMenu(None)

    def _check_notifications(self):
        self.clear_notifications()
        if self._specification is not None:
            report = self._specification.settings.report_inconsistencies()
            for message in report:
                self.add_notification(message)



######################################################################################################################
# Copyright (C) 2017-2023 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains utilities for Exporter."""
from dataclasses import dataclass
from spine_engine.project_item.project_item_resource import url_resource
from spine_items.utils import convert_to_sqlalchemy_url

EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX = ".export-manifest"
"""Prefix for the temporary files that exporter's executable uses to communicate output paths."""


@dataclass
class Database:
    """Legacy class for Database specific export settings."""

    url: str = ""
    """Database URL."""
    output_file_name: str = ""
    """Output file name; relative to item's data dir."""

    @staticmethod
    def from_dict(database_dict):
        """
        Deserializes :class:`Database` from a dictionary.

        Args:
            database_dict (dict): serialized :class:`Database`

        Returns:
            Database: deserialized instance
        """
        db = Database()
        db.output_file_name = database_dict["output_file_name"]
        return db


def output_database_resources(item_name, output_channels):
    """Gathers output database resources from output channels that have an out URL set.

    Args
        item_name (str): exporter's name
        output_channels (Iterable of OutputChannel): output channels

    Returns:
        list of ProjectItemResource: database resources
    """
    resources = []
    for channel in output_channels:
        if channel.out_url is None:
            continue
        url = str(convert_to_sqlalchemy_url(channel.out_url))
        resources.append(url_resource(item_name, url, channel.out_label))
    return resources



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for DataConnectionIcon class."""
from tempfile import TemporaryDirectory
import unittest
from PySide6.QtWidgets import QApplication
from spine_items.data_connection.data_connection_factory import DataConnectionFactory
from tests.mock_helpers import clean_up_toolbox, create_toolboxui_with_project


class TestDataConnectionIcon(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        super().setUp()
        self._temp_dir = TemporaryDirectory()
        self._toolbox = create_toolboxui_with_project(self._temp_dir.name)
        item_dict = {"type": "Data Connection", "description": "", "x": 0, "y": 0}
        dc = DataConnectionFactory.make_item("DC", item_dict, self._toolbox, self._toolbox.project())
        self._toolbox.project().add_item(dc)

    def tearDown(self):
        super().tearDown()
        clean_up_toolbox(self._toolbox)
        self._temp_dir.cleanup()

    def test_make_data_connection_icon(self):
        _ = self._toolbox.project()._project_items["DC"].get_icon()


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for Import editor's SourceDataTableModel."""
import unittest
from unittest.mock import MagicMock
from PySide6.QtCore import Qt
from spine_items.importer.mvcmodels.mappings_model import MappingsModel
from spine_items.importer.mvcmodels.mappings_model_roles import Role
from spine_items.importer.mvcmodels.source_data_table_model import SourceDataTableModel
from spinedb_api.import_mapping.import_mapping_compat import import_mapping_from_dict
from spinedb_api.import_mapping.type_conversion import value_to_convert_spec


class TestSourceDataTableModel(unittest.TestCase):
    def setUp(self):
        self._model = SourceDataTableModel()

    def tearDown(self):
        self._model.deleteLater()

    def test_column_type_checking(self):
        self._model.reset_model([["1", "0h", "2018-01-01 00:00"], ["2", "1h", "2018-01-01 00:00"]])
        self._model.set_type(0, value_to_convert_spec("float"))
        self.assertEqual(self._model._column_type_errors, {})
        self.assertEqual(self._model._row_type_errors, {})
        self._model.set_type(1, value_to_convert_spec("duration"))
        self.assertEqual(self._model._column_type_errors, {})
        self.assertEqual(self._model._row_type_errors, {})
        self._model.set_type(2, value_to_convert_spec("datetime"))
        self.assertEqual(self._model._column_type_errors, {})
        self.assertEqual(self._model._row_type_errors, {})

    def test_row_type_checking(self):
        self._model.reset_model(
            [["1", "1", "1.1"], ["2h", "1h", "2h"], ["2018-01-01 00:00", "2018-01-01 00:00", "2018-01-01 00:00"]]
        )
        self._model.set_type(0, value_to_convert_spec("float"), orientation=Qt.Orientation.Vertical)
        self.assertEqual(self._model._column_type_errors, {})
        self.assertEqual(self._model._row_type_errors, {})
        self._model.set_type(1, value_to_convert_spec("duration"), orientation=Qt.Orientation.Vertical)
        self.assertEqual(self._model._column_type_errors, {})
        self.assertEqual(self._model._row_type_errors, {})
        self._model.set_type(2, value_to_convert_spec("datetime"), orientation=Qt.Orientation.Vertical)
        self.assertEqual(self._model._column_type_errors, {})
        self.assertEqual(self._model._row_type_errors, {})

    def test_column_type_checking_produces_error(self):
        self._model.reset_model([["Not a valid number", "2.4"], ["1", "3"]])
        self._model.set_type(0, value_to_convert_spec("float"))
        error_index = (0, 0)
        self.assertEqual(len(self._model._column_type_errors), 1)
        self.assertEqual(self._model._row_type_errors, {})
        self.assertTrue(error_index in self._model._column_type_errors)
        self.assertEqual(self._model.data(self._model.index(*error_index)), "Not a valid number")

        # if we add a pivoted mapping for the row with the error, the error should not be
        undo_stack = MagicMock()
        mappings_model = MappingsModel(undo_stack, None)
        list_index = self._add_mapping(
            mappings_model, {"map_type": "ObjectClass", "name": {"map_type": "row", "value_reference": 0}}
        )
        self._model.set_mapping_list_index(list_index)
        self.assertEqual(self._model.data(self._model.index(*error_index)), "Not a valid number")

        # or if we add a mapping where there reading starts from a row below the error, the error should not be shown.
        list_index = self._add_mapping(mappings_model, {"map_type": "ObjectClass", "read_start_row": 1})
        self._model.set_mapping_list_index(list_index)
        self.assertEqual(self._model.data(self._model.index(*error_index)), "Not a valid number")
        mappings_model.deleteLater()

    def test_row_type_checking_produces_error(self):
        self._model.reset_model([["1", "2.4"], ["Not a valid number", "3"]])
        self._model.set_type(1, value_to_convert_spec("float"), orientation=Qt.Orientation.Vertical)
        error_index = (1, 0)
        self.assertEqual(len(self._model._row_type_errors), 1)
        self.assertEqual(self._model._column_type_errors, {})
        self.assertTrue(error_index in self._model._row_type_errors)
        # Error should only be shown if we have a pivot mapping on that row.
        self.assertEqual(self._model.data(self._model.index(*error_index)), "Not a valid number")

        # if we add mapping error should be shown.
        undo_stack = MagicMock()
        mappings_model = MappingsModel(undo_stack, None)
        list_index = self._add_mapping(
            mappings_model, {"map_type": "ObjectClass", "name": {"map_type": "row", "value_reference": 1}}
        )
        self._model.set_mapping_list_index(list_index)
        self.assertEqual(self._model.data(self._model.index(*error_index)), "Not a valid number")
        mappings_model.deleteLater()

    def test_mapping_column_colors(self):
        self._model.reset_model([[1, 2], [3, 4]])
        # column mapping
        undo_stack = MagicMock()
        mappings_model = MappingsModel(undo_stack, None)
        list_index = self._add_mapping(mappings_model, {"map_type": "ObjectClass", "name": 0})
        self._model.set_mapping_list_index(list_index)
        entity_class_color = self._find_color(list_index, "Entity class names")
        self.assertEqual(
            self._model.data(self._model.index(0, 0), role=Qt.ItemDataRole.BackgroundRole), entity_class_color
        )
        self.assertEqual(
            self._model.data(self._model.index(1, 0), role=Qt.ItemDataRole.BackgroundRole), entity_class_color
        )
        # row not showing color if the start reading row is specified
        list_index = self._add_mapping(mappings_model, {"map_type": "ObjectClass", "name": 0, "read_start_row": 1})
        self._model.set_mapping_list_index(list_index)
        entity_class_color = self._find_color(list_index, "Entity class names")
        self.assertEqual(self._model.data(self._model.index(0, 0), role=Qt.ItemDataRole.BackgroundRole), None)
        self.assertEqual(
            self._model.data(self._model.index(1, 0), role=Qt.ItemDataRole.BackgroundRole), entity_class_color
        )
        # row not showing color if the row is pivoted
        list_index = self._add_mapping(
            mappings_model, {"map_type": "ObjectClass", "name": 0, "object": {"map_type": "row", "value_reference": 0}}
        )
        self._model.set_mapping_list_index(list_index)
        self.assertNotEqual(
            self._model.data(self._model.index(0, 0), role=Qt.ItemDataRole.BackgroundRole), entity_class_color
        )
        self.assertEqual(
            self._model.data(self._model.index(1, 0), role=Qt.ItemDataRole.BackgroundRole), entity_class_color
        )
        mappings_model.deleteLater()

    def test_mapping_pivoted_colors(self):
        self._model.reset_model([[1, 2], [3, 4]])
        # row mapping
        undo_stack = MagicMock()
        mappings_model = MappingsModel(undo_stack, None)
        list_index = self._add_mapping(
            mappings_model, {"map_type": "ObjectClass", "object": {"map_type": "row", "value_reference": 0}}
        )
        self._model.set_mapping_list_index(list_index)
        entity_color = self._find_color(list_index, "Entity names")
        metadata_color = self._find_color(list_index, "Entity metadata")
        self.assertEqual(self._model.data(self._model.index(0, 0), role=Qt.ItemDataRole.BackgroundRole), entity_color)
        self.assertEqual(self._model.data(self._model.index(0, 1), role=Qt.ItemDataRole.BackgroundRole), entity_color)
        self.assertEqual(self._model.data(self._model.index(1, 0), role=Qt.ItemDataRole.BackgroundRole), metadata_color)
        self.assertEqual(self._model.data(self._model.index(1, 1), role=Qt.ItemDataRole.BackgroundRole), metadata_color)
        # column not showing color if the columns is skipped
        list_index = self._add_mapping(
            mappings_model,
            {"map_type": "ObjectClass", "object": {"map_type": "row", "value_reference": 0}, "skip_columns": [0]},
        )
        self._model.set_mapping_list_index(list_index)
        entity_color = self._find_color(list_index, "Entity names")
        self.assertEqual(self._model.data(self._model.index(0, 0), role=Qt.ItemDataRole.BackgroundRole), None)
        self.assertEqual(self._model.data(self._model.index(0, 1), role=Qt.ItemDataRole.BackgroundRole), entity_color)
        self.assertEqual(self._model.data(self._model.index(1, 0), role=Qt.ItemDataRole.BackgroundRole), None)
        mappings_model.deleteLater()

    def test_mapping_column_and_pivot_colors(self):
        self._model.reset_model([[1, 2], [3, 4]])
        # row mapping
        undo_stack = MagicMock()
        mappings_model = MappingsModel(undo_stack, None)
        list_index = self._add_mapping(
            mappings_model, {"map_type": "ObjectClass", "name": 0, "object": {"map_type": "row", "value_reference": 0}}
        )
        self._model.set_mapping_list_index(list_index)
        entity_class_color = self._find_color(list_index, "Entity class names")
        entity_color = self._find_color(list_index, "Entity names")
        metadata_color = self._find_color(list_index, "Entity metadata")
        self.assertEqual(self._model.data(self._model.index(0, 0), role=Qt.ItemDataRole.BackgroundRole), None)
        self.assertEqual(self._model.data(self._model.index(0, 1), role=Qt.ItemDataRole.BackgroundRole), entity_color)
        self.assertEqual(
            self._model.data(self._model.index(1, 0), role=Qt.ItemDataRole.BackgroundRole), entity_class_color
        )
        self.assertEqual(self._model.data(self._model.index(1, 1), role=Qt.ItemDataRole.BackgroundRole), metadata_color)
        mappings_model.deleteLater()

    def test_mapping_column_and_pivot_colors_with_value_mapping_position_set_to_random_column(self):
        self._model.reset_model([[1, 2, 3, 4], [4, 5, 6, 7]])
        # row mapping
        undo_stack = MagicMock()
        mappings_model = MappingsModel(undo_stack, None)
        list_index = self._add_mapping(
            mappings_model,
            [
                {"map_type": "ObjectClass", "position": "hidden", "value": "timeline"},
                {"map_type": "Object", "position": 1},
                {"map_type": "ObjectMetadata", "position": "hidden"},
                {"map_type": "ParameterDefinition", "position": -1},
                {"map_type": "Alternative", "position": 0},
                {"map_type": "ParameterValueMetadata", "position": "hidden"},
                {"map_type": "ParameterValueType", "position": "hidden", "value": "map"},
                {"map_type": "IndexName", "position": "hidden", "value": "timestep"},
                {"map_type": "ParameterValueIndex", "position": 2},
                {"map_type": "ExpandedValue", "position": 1},  # This caused a failure in the past.
            ],
        )
        self._model.set_mapping_list_index(list_index)
        # no color showing where row and column mapping intersect
        entity_color = self._find_color(list_index, "Entity names")
        parameter_definition_color = self._find_color(list_index, "Parameter names")
        alternative_color = self._find_color(list_index, "Alternative names")
        index_color = self._find_color(list_index, "Parameter indexes")
        value_color = self._find_color(list_index, "Parameter values")
        self.assertEqual(self._model.data(self._model.index(0, 0), role=Qt.ItemDataRole.BackgroundRole), None)
        self.assertEqual(self._model.data(self._model.index(0, 1), role=Qt.ItemDataRole.BackgroundRole), None)
        self.assertEqual(self._model.data(self._model.index(0, 2), role=Qt.ItemDataRole.BackgroundRole), None)
        self.assertEqual(
            self._model.data(self._model.index(0, 3), role=Qt.ItemDataRole.BackgroundRole), parameter_definition_color
        )
        self.assertEqual(
            self._model.data(self._model.index(1, 0), role=Qt.ItemDataRole.BackgroundRole), alternative_color
        )
        self.assertEqual(self._model.data(self._model.index(1, 1), role=Qt.ItemDataRole.BackgroundRole), entity_color)
        self.assertEqual(self._model.data(self._model.index(1, 2), role=Qt.ItemDataRole.BackgroundRole), index_color)
        self.assertEqual(self._model.data(self._model.index(1, 3), role=Qt.ItemDataRole.BackgroundRole), value_color)
        mappings_model.deleteLater()

    def test_mapping_column_and_pivot_colors_with_value_mapping_position_set_to_column_in_pivot(self):
        self._model.reset_model([[1, 2, 3, 4], [4, 5, 6, 7]])
        # row mapping
        undo_stack = MagicMock()
        mappings_model = MappingsModel(undo_stack, None)
        list_index = self._add_mapping(
            mappings_model,
            [
                {"map_type": "ObjectClass", "position": "hidden", "value": "timeline"},
                {"map_type": "Object", "position": 1},
                {"map_type": "ObjectMetadata", "position": "hidden"},
                {"map_type": "ParameterDefinition", "position": -1},
                {"map_type": "Alternative", "position": 0},
                {"map_type": "ParameterValueMetadata", "position": "hidden"},
                {"map_type": "ParameterValueType", "position": "hidden", "value": "map"},
                {"map_type": "IndexName", "position": "hidden", "value": "timestep"},
                {"map_type": "ParameterValueIndex", "position": 2},
                {"map_type": "ExpandedValue", "position": 3},  # This caused a failure in the past.
            ],
        )
        self._model.set_mapping_list_index(list_index)
        # no color showing where row and column mapping intersect
        entity_color = self._find_color(list_index, "Entity names")
        parameter_definition_color = self._find_color(list_index, "Parameter names")
        alternative_color = self._find_color(list_index, "Alternative names")
        index_color = self._find_color(list_index, "Parameter indexes")
        value_color = self._find_color(list_index, "Parameter values")
        self.assertEqual(self._model.data(self._model.index(0, 0), role=Qt.ItemDataRole.BackgroundRole), None)
        self.assertEqual(self._model.data(self._model.index(0, 1), role=Qt.ItemDataRole.BackgroundRole), None)
        self.assertEqual(self._model.data(self._model.index(0, 2), role=Qt.ItemDataRole.BackgroundRole), None)
        self.assertEqual(
            self._model.data(self._model.index(0, 3), role=Qt.ItemDataRole.BackgroundRole), parameter_definition_color
        )
        self.assertEqual(
            self._model.data(self._model.index(1, 0), role=Qt.ItemDataRole.BackgroundRole), alternative_color
        )
        self.assertEqual(self._model.data(self._model.index(1, 1), role=Qt.ItemDataRole.BackgroundRole), entity_color)
        self.assertEqual(self._model.data(self._model.index(1, 2), role=Qt.ItemDataRole.BackgroundRole), index_color)
        self.assertEqual(self._model.data(self._model.index(1, 3), role=Qt.ItemDataRole.BackgroundRole), value_color)
        mappings_model.deleteLater()

    @staticmethod
    def _find_color(list_index, item_name):
        flattened_mappings = list_index.data(Role.FLATTENED_MAPPINGS)
        row = flattened_mappings.display_names.index(item_name)
        return flattened_mappings.display_colors[row]

    @staticmethod
    def _add_mapping(mappings_model, mapping_dict):
        mapping = import_mapping_from_dict(mapping_dict)
        mappings_model.append_new_table_with_mapping(f"source table {mappings_model.rowCount()}", mapping)
        table_index = mappings_model.index(mappings_model.rowCount() - 1, 0)
        return mappings_model.index(0, 0, table_index)


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Exporter project item info."""
from spine_engine.project_item.project_item_info import ProjectItemInfo


class ItemInfo(ProjectItemInfo):
    @staticmethod
    def item_type():
        """See base class."""
        return "Exporter"



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Classes for context menus used alongside the Importer project item."""
from PySide6.QtCore import QPoint, Signal, Slot
from PySide6.QtWidgets import QMenu
from spinetoolbox.mvcmodels.filter_checkbox_list_model import DataToValueFilterCheckboxListModel
from spinetoolbox.widgets.custom_menus import CustomContextMenu, FilterMenuBase, ItemSpecificationMenu
from ..mvcmodels.mappings_model_roles import Role


class FilesContextMenu(CustomContextMenu):
    """Context menu class for source files view in Importer properties tab."""

    def __init__(self, parent, position, index):
        """
        Args:
            parent (QWidget): Parent for menu widget (ToolboxUI)
            position (QPoint): Position on screen
            index (QModelIndex): Index of item that requested the context-menu
        """
        super().__init__(parent, position)
        if not index.isValid():
            self.add_action("Open directory...")
        else:
            self.add_action("Open import editor...")
            self.addSeparator()
            self.add_action("Open directory...")


class SpecificationMenu(ItemSpecificationMenu):
    """Context menu class for Data transformer specifications."""


class SourceListMenu(CustomContextMenu):
    """
    Menu for source list.
    """

    def __init__(self, parent, position, can_paste_option, can_paste_mapping):
        super().__init__(parent, position)
        self.add_action("Copy options")
        self.add_action("Copy mappings")
        self.add_action("Copy options and mappings")
        self.addSeparator()
        self.add_action("Paste options", enabled=can_paste_option)
        self.add_action("Paste mappings", enabled=can_paste_mapping)
        self.add_action("Paste options and mappings", enabled=can_paste_mapping & can_paste_option)


class MappingListMenu(CustomContextMenu):
    """
    Menu for source list.
    """

    def __init__(self, parent, position, can_copy, can_paste):
        super().__init__(parent, position)
        self.add_action("Copy mapping(s)", enabled=can_copy)
        self.add_action("Paste mapping(s)", enabled=can_paste)


class SourceDataTableMenu(QMenu):
    """
    A context menu for the source data table, to let users define a Mapping from a data table.
    """

    def __init__(self, model, ui):
        """
        Args:
            model (MappingsModel): model
            ui (Any): import editor's UI
        """
        super().__init__(ui.source_data_table)
        self._model = model
        self._ui = ui

    def _set_mapping(self, flattened_mappings, index, map_type=None, value=None):
        self._model.change_component_mapping(flattened_mappings, index, map_type, value)

    @Slot(QPoint)
    def request_menu(self, pos=None):
        indexes = self.parent().selectedIndexes()
        if not indexes:
            return
        self.clear()
        index = indexes[0]
        row = index.row() + 1
        col = index.column() + 1

        list_index = self._ui.mapping_list.selectionModel().currentIndex()
        if list_index.isValid():
            mapping_names = [
                self._model.index(row, 0, list_index).data() for row in range(self._model.rowCount(list_index))
            ]
            flattened_mappings = list_index.data(Role.FLATTENED_MAPPINGS)
        else:
            mapping_names = None
            flattened_mappings = None

        def create_callback(row, map_type, value):
            component_index = self._model.index(row, 0, list_index)
            return lambda: self._set_mapping(flattened_mappings, component_index, map_type, value)

        menus = [
            ("Map column to...", "Column", col),
            ("Map header to...", "Column Header", col),
            ("Map row to...", "Row", row),
            ("Map all headers to...", "Headers", 0),
        ]

        for title, map_type, value in menus:
            m = self.addMenu(title)
            if mapping_names is not None:
                for row, name in enumerate(mapping_names):
                    m.addAction(name).triggered.connect(create_callback(row, map_type, value))
            else:
                m.addAction("<no mapping selected>").setEnabled(False)

        global_pos = self.parent().mapToGlobal(QPoint(5, 20))
        menu_pos = global_pos + pos
        self.move(menu_pos)
        self.show()


class SimpleFilterMenu(FilterMenuBase):
    filterChanged = Signal(set)

    def __init__(self, parent, show_empty=True):
        """
        Args:
            parent (SpineDBEditor)
        """
        super().__init__(parent)
        self._set_up(DataToValueFilterCheckboxListModel, self, str, show_empty=show_empty)

    def emit_filter_changed(self, valid_values):
        self.filterChanged.emit(valid_values)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for Data Connection project item."""
import gc
import os
from pathlib import Path
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from unittest.mock import MagicMock, NonCallableMagicMock
from PySide6.QtCore import QItemSelectionModel
from PySide6.QtGui import Qt
from PySide6.QtWidgets import QApplication, QDialog, QMessageBox
from spine_items.data_connection.data_connection import DataConnection, _Role
from spine_items.data_connection.data_connection_factory import DataConnectionFactory
from spine_items.data_connection.item_info import ItemInfo
from spinedb_api import create_new_spine_database
from spinetoolbox.helpers import signal_waiter
from tests.mock_helpers import (
    clean_up_toolbox,
    create_mock_project,
    create_mock_toolbox,
    create_toolboxui_with_project,
    mock_finish_project_item_construction,
)


class TestDataConnection(unittest.TestCase):
    def test_item_type(self):
        self.assertEqual(DataConnection.item_type(), ItemInfo.item_type())


class TestDataConnectionWithProject(unittest.TestCase):
    def setUp(self):
        """Set up."""
        self._temp_dir = TemporaryDirectory()
        self._toolbox = create_toolboxui_with_project(self._temp_dir.name)
        factory = DataConnectionFactory()
        item_dict = {"type": "Data Connection", "description": "", "db_references": [], "x": 0, "y": 0}
        project = self._toolbox.project()
        self._data_connection = factory.make_item("DC", item_dict, self._toolbox, project)
        project.add_item(self._data_connection)
        self._ref_model = self._data_connection.reference_model

    def tearDown(self):
        clean_up_toolbox(self._toolbox)
        gc.collect()
        self._temp_dir.cleanup()

    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_add_file_references(self):
        temp_dir = Path(self._temp_dir.name, "references")
        temp_dir.mkdir()
        with mock.patch("spine_items.data_connection.data_connection.QFileDialog.getOpenFileNames") as mock_filenames:
            a = Path(temp_dir, "a.txt")
            a.touch()
            b = Path(temp_dir, "b.txt")
            b.touch()
            c = Path(temp_dir, "c.txt")  # Note. Does not exist
            # Add nothing
            mock_filenames.return_value = ([], "*.*")
            self._data_connection.show_add_file_references_dialog()
            self.assertEqual(1, mock_filenames.call_count)
            self.assertEqual(0, len(self._data_connection.file_references))
            self.assertEqual(0, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            # Add one file
            mock_filenames.return_value = ([str(a)], "*.*")
            self._data_connection.show_add_file_references_dialog()
            self.assertEqual(2, mock_filenames.call_count)
            self.assertEqual(1, len(self._data_connection.file_references))
            self.assertEqual(1, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            self.assertEqual(str(a), self._ref_model.index(0, 0, self._ref_model.index(0, 0)).data())
            # Try to add a path that has already been added
            self._data_connection.show_add_file_references_dialog()
            self.assertEqual(3, mock_filenames.call_count)
            self.assertEqual(1, len(self._data_connection.file_references))
            self.assertEqual(1, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            self.assertEqual(str(a), self._ref_model.index(0, 0, self._ref_model.index(0, 0)).data())
            self._data_connection.file_references = []
            self._data_connection.populate_reference_list([])
            # Add two references (the other one is non-existing)
            # Note: non-existing files cannot be added with the toolbox but this tests a situation when
            # project.json file has references to files that do not exist anymore and user tries to add a
            # new reference to a Data Connection that contains non-existing file references
            mock_filenames.return_value = ([str(b), str(c)], "*.*")
            self._data_connection.show_add_file_references_dialog()
            self.assertEqual(4, mock_filenames.call_count)
            self.assertEqual(1, len(self._data_connection.file_references))
            self.assertEqual(1, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            self.assertEqual(str(b), self._ref_model.index(0, 0, self._ref_model.index(0, 0)).data())
            # Now add new reference
            mock_filenames.return_value = ([str(a)], "*.*")
            self._data_connection.show_add_file_references_dialog()
            self.assertEqual(5, mock_filenames.call_count)
            self.assertEqual(2, len(self._data_connection.file_references))
            self.assertEqual(2, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            self.assertEqual(str(b), self._ref_model.index(0, 0, self._ref_model.index(0, 0)).data())
            self.assertEqual(str(a), self._ref_model.index(1, 0, self._ref_model.index(0, 0)).data())

    def test_add_db_references(self):
        with (
            mock.patch("spine_items.data_connection.data_connection.UrlSelectorDialog.exec") as url_selector_exec,
            mock.patch(
                "spine_items.data_connection.data_connection.UrlSelectorDialog.url_dict"
            ) as url_selector_url_dict,
        ):
            # Add nothing
            url_selector_exec.return_value = QDialog.DialogCode.Rejected
            self._data_connection.show_add_db_reference_dialog()
            self.assertEqual(1, url_selector_exec.call_count)
            self.assertFalse(self._data_connection.has_db_references())
            self.assertEqual(0, self._ref_model.rowCount(self._ref_model.index(1, 0)))
            # Add one url
            url_selector_exec.return_value = QDialog.DialogCode.Accepted
            url_selector_url_dict.return_value = {
                "dialect": "mysql",
                "host": "host",
                "port": 3306,
                "database": "db",
                "username": "randy",
                "password": "creamfraiche",
            }
            self._data_connection.show_add_db_reference_dialog()
            self.assertEqual(2, url_selector_exec.call_count)
            self.assertEqual(1, len(list(self._data_connection.db_reference_iter())))
            self.assertEqual(1, self._ref_model.rowCount(self._ref_model.index(1, 0)))
            # Add same url with different username and password (should not be added)
            url_selector_url_dict.return_value = {
                "dialect": "mysql",
                "host": "host",
                "port": 3306,
                "database": "db",
                "username": "scott",
                "password": "tiger",
            }
            self._data_connection.show_add_db_reference_dialog()
            self.assertEqual(3, url_selector_exec.call_count)
            self.assertEqual(1, len(list(self._data_connection.db_reference_iter())))
            self.assertEqual(1, self._ref_model.rowCount(self._ref_model.index(1, 0)))

    def test_remove_references(self):
        temp_dir = Path(self._temp_dir.name, "references")
        temp_dir.mkdir()
        with (
            mock.patch("spine_items.data_connection.data_connection.QFileDialog.getOpenFileNames") as mock_filenames,
            mock.patch.object(
                self._data_connection._properties_ui.treeView_dc_references, "selectedIndexes"
            ) as mock_selected_indexes,
            mock.patch("spine_items.data_connection.data_connection.UrlSelectorDialog.exec") as url_selector_exec,
            mock.patch(
                "spine_items.data_connection.data_connection.UrlSelectorDialog.url_dict"
            ) as url_selector_url_dict,
        ):
            a = Path(temp_dir, "a.txt")
            a.touch()
            b = Path(temp_dir, "b.txt")
            b.touch()
            c = Path(temp_dir, "c.txt")  # Note. This file is not actually created
            d = Path(temp_dir, "d.txt")  # Note. This file is not actually created
            self.assertTrue(os.path.isfile(str(a)) and os.path.isfile(str(b)))  # existing files
            self.assertFalse(os.path.isfile(str(c)))  # non-existing file
            self.assertFalse(os.path.isfile(str(d)))  # non-existing file
            # First add a couple of files as refs
            mock_filenames.return_value = ([str(a), str(b)], "*.*")
            self._data_connection.show_add_file_references_dialog()
            self.assertEqual(1, mock_filenames.call_count)
            self.assertEqual(2, len(self._data_connection.file_references))
            self.assertEqual(2, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            # Second add a couple of dbs as refs
            url_selector_exec.return_value = QDialog.DialogCode.Accepted
            url_selector_url_dict.return_value = {
                "dialect": "mysql",
                "username": "scott",
                "password": "tiger",
                "host": "host",
                "port": 3306,
                "database": "db",
            }
            self._data_connection.show_add_db_reference_dialog()
            url_selector_url_dict.return_value = {
                "dialect": "mysql",
                "username": "randy",
                "password": "creamfraiche",
                "host": "host",
                "port": 3307,
                "database": "db",
            }
            self._data_connection.show_add_db_reference_dialog()
            self.assertEqual(2, url_selector_exec.call_count)
            self.assertEqual(2, len(list(self._data_connection.db_reference_iter())))
            self.assertEqual(2, self._ref_model.rowCount(self._ref_model.index(1, 0)))
            # Test with no indexes selected
            mock_selected_indexes.return_value = []
            self._data_connection.remove_references()
            self.assertEqual(1, mock_selected_indexes.call_count)
            self.assertEqual(2, len(self._data_connection.file_references))
            self.assertEqual(2, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            # Set one file selected and remove it
            a_index = self._ref_model.index(0, 0, self._ref_model.index(0, 0))
            mock_selected_indexes.return_value = [a_index]
            self._data_connection.remove_references()
            self.assertEqual(2, mock_selected_indexes.call_count)
            self.assertEqual(1, len(self._data_connection.file_references))
            self.assertEqual(1, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            # Check that the remaining file is the one that's supposed to be there
            self.assertEqual([str(b)], self._data_connection.file_references)
            self.assertEqual(str(b), self._ref_model.item(0).child(0).data(Qt.ItemDataRole.DisplayRole))
            # Set one db selected and remove it
            db1_index = self._ref_model.index(0, 0, self._ref_model.index(1, 0))
            mock_selected_indexes.return_value = [db1_index]
            self._data_connection.remove_references()
            self.assertEqual(3, mock_selected_indexes.call_count)
            self.assertEqual(1, len(list(self._data_connection.db_reference_iter())))
            self.assertEqual(1, self._ref_model.rowCount(self._ref_model.index(1, 0)))
            # Check that the remaining db is the one that's supposed to be there
            self.assertEqual(
                [
                    {
                        "dialect": "mysql",
                        "host": "host",
                        "port": 3307,
                        "database": "db",
                        "username": "randy",
                        "password": "creamfraiche",
                    }
                ],
                list(self._data_connection.db_reference_iter()),
            )
            self.assertEqual(
                "mysql+pymysql://host:3307/db", self._ref_model.item(1).child(0).data(Qt.ItemDataRole.DisplayRole)
            )
            # Now remove the remaining file and db
            b_index = self._ref_model.index(0, 0, self._ref_model.index(0, 0))
            db2_index = self._ref_model.index(0, 0, self._ref_model.index(1, 0))
            mock_selected_indexes.return_value = [b_index, db2_index]
            self._data_connection.remove_references()
            self.assertEqual(4, mock_selected_indexes.call_count)
            self.assertEqual(0, len(self._data_connection.file_references))
            self.assertEqual(0, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            self.assertEqual(0, len(list(self._data_connection.db_reference_iter())))
            self.assertEqual(0, self._ref_model.rowCount(self._ref_model.index(1, 0)))
            # Add a and b back and the two non-existing files as well
            # Select non-existing file c and remove it
            mock_filenames.return_value = ([str(a), str(b), str(c), str(d)], "*.*")
            self._data_connection.show_add_file_references_dialog()
            self.assertEqual(2, mock_filenames.call_count)
            self.assertEqual(2, len(self._data_connection.file_references))
            self.assertEqual(2, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            # Check that the two remaining items are the ones that are supposed to be there
            self.assertEqual(str(a), self._data_connection.file_references[0])
            self.assertEqual(str(a), self._ref_model.item(0).child(0).data(Qt.ItemDataRole.DisplayRole))
            self.assertEqual(str(b), self._data_connection.file_references[1])
            self.assertEqual(str(b), self._ref_model.item(0).child(1).data(Qt.ItemDataRole.DisplayRole))
            # Now select the two remaining ones and remove them
            a_index = self._ref_model.index(0, 0, self._ref_model.index(0, 0))
            b_index = self._ref_model.index(1, 0, self._ref_model.index(0, 0))
            mock_selected_indexes.return_value = [a_index, b_index]
            self._data_connection.remove_references()
            self.assertEqual(5, mock_selected_indexes.call_count)
            self.assertEqual(0, len(self._data_connection.file_references))
            self.assertEqual(0, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            # Add a, b, c, and d back Select all and remove.
            mock_filenames.return_value = ([str(a), str(b), str(c), str(d)], "*.*")
            self._data_connection.show_add_file_references_dialog()
            self.assertEqual(3, mock_filenames.call_count)
            a_index = self._ref_model.index(0, 0, self._ref_model.index(0, 0))
            b_index = self._ref_model.index(1, 0, self._ref_model.index(0, 0))
            mock_selected_indexes.return_value = [a_index, b_index]
            self._data_connection.remove_references()
            self.assertEqual(6, mock_selected_indexes.call_count)
            self.assertEqual(0, len(self._data_connection.file_references))
            self.assertEqual(0, self._ref_model.rowCount(self._ref_model.index(0, 0)))

    def test_remove_references_with_del_key(self):
        temp_dir = Path(self._temp_dir.name, "references")
        temp_dir.mkdir()
        with (
            mock.patch("spine_items.data_connection.data_connection.QFileDialog.getOpenFileNames") as mock_filenames,
            mock.patch("spine_items.data_connection.data_connection.UrlSelectorDialog.exec"),
            mock.patch(
                "spine_items.data_connection.data_connection.UrlSelectorDialog.url", new_callable=mock.PropertyMock
            ) as url_selector_url,
        ):
            a = Path(temp_dir, "a.txt")
            a.touch()
            b = Path(temp_dir, "b.txt")
            b.touch()
            self.assertTrue(os.path.isfile(str(a)) and os.path.isfile(str(b)))  # existing files
            self._data_connection.restore_selections()
            self._data_connection._connect_signals()
            # First add a couple of files as refs
            mock_filenames.return_value = ([str(a), str(b)], "*.*")
            self._data_connection.show_add_file_references_dialog()
            self.assertEqual(1, mock_filenames.call_count)
            self.assertEqual(2, len(self._data_connection.file_references))
            self.assertEqual(2, self._ref_model.rowCount(self._ref_model.index(0, 0)))
            # Second add a couple of dbs as refs
            url_selector_url.return_value = "mysql://scott:tiger@host:3306/db"
            self._data_connection.show_add_db_reference_dialog()
            indexes = self._data_connection._properties_ui.treeView_dc_references.selectedIndexes()
            self.assertTrue(len(indexes) == 0)
            # Set index selected
            file_ref_root_index = self._ref_model.index(0, 0)
            ref_index = self._ref_model.index(0, 0, file_ref_root_index)
            self._data_connection._properties_ui.treeView_dc_references.selectionModel().select(
                ref_index, QItemSelectionModel.Select
            )
            indexes = self._data_connection._properties_ui.treeView_dc_references.selectedIndexes()
            self.assertTrue(len(indexes) == 1)
            self._data_connection._properties_ui.treeView_dc_references.del_key_pressed.emit()
            self.assertEqual(1, len(self._data_connection.file_references))
            # Remove remaining two simultaneously by selecting bith and removing them with delete key
            file_ref_index = self._ref_model.index(0, 0, self._ref_model.index(0, 0))
            db_ref_index = self._ref_model.index(0, 0, self._ref_model.index(1, 0))
            self._data_connection._properties_ui.treeView_dc_references.selectionModel().select(
                file_ref_index, QItemSelectionModel.Select
            )
            self._data_connection._properties_ui.treeView_dc_references.selectionModel().select(
                db_ref_index, QItemSelectionModel.Select
            )
            indexes = self._data_connection._properties_ui.treeView_dc_references.selectedIndexes()
            self._data_connection._properties_ui.treeView_dc_references.del_key_pressed.emit()
            self.assertEqual(0, len(self._data_connection.file_references))
            self.assertEqual(0, len(list(self._data_connection.db_reference_iter())))

    def test_renaming_file_marks_its_reference_as_missing(self):
        temp_dir = Path(self._temp_dir.name, "references")
        temp_dir.mkdir()
        a = Path(temp_dir, "a.txt")
        a.touch()
        with mock.patch("spine_items.data_connection.data_connection.QFileDialog.getOpenFileNames") as mock_filenames:
            mock_filenames.return_value = ([str(a)], "*.*")
            self._data_connection.show_add_file_references_dialog()
        renamed_file = a.parent / "renamed.txt"
        with signal_waiter(self._data_connection.file_system_watcher.file_renamed) as waiter:
            a.rename(renamed_file)
            waiter.wait()
        index = self._ref_model.index(0, 0, self._ref_model.index(0, 0))
        self.assertEqual(index.data(), str(a))
        self.assertTrue(index.data(Qt.ItemDataRole.UserRole + 2))
        self.assertEqual(self._data_connection.file_references, [str(a)])

    def test_deleting_file_marks_its_reference_as_missing(self):
        temp_dir = Path(self._temp_dir.name, "references")
        temp_dir.mkdir()
        a = Path(temp_dir, "a.txt")
        a.touch()
        with mock.patch("spine_items.data_connection.data_connection.QFileDialog.getOpenFileNames") as mock_filenames:
            mock_filenames.return_value = ([str(a)], "*.*")
            self._data_connection.show_add_file_references_dialog()
        with signal_waiter(self._data_connection.file_system_watcher.file_removed) as waiter:
            a.unlink()
            waiter.wait()
        index = self._ref_model.index(0, 0, self._ref_model.index(0, 0))
        self.assertEqual(index.data(), str(a))
        self.assertTrue(index.data(Qt.ItemDataRole.UserRole + 2))
        self.assertEqual(self._data_connection.file_references, [str(a)])

    def test_copy_reference_to_project(self):
        temp_dir = Path(self._temp_dir.name, "references")
        temp_dir.mkdir()
        a = Path(temp_dir, "a.txt")
        a.touch()
        with mock.patch("spine_items.data_connection.data_connection.QFileDialog.getOpenFileNames") as mock_filenames:
            mock_filenames.return_value = ([str(a)], "*.*")
            self._data_connection.show_add_file_references_dialog()
        self._data_connection.restore_selections()
        ref_root_index = self._ref_model.index(0, 0)
        ref_index = self._ref_model.index(0, 0, ref_root_index)
        properties_ui = self._toolbox.project_item_properties_ui(self._data_connection.item_type())
        properties_ui.treeView_dc_references.selectionModel().select(ref_index, QItemSelectionModel.ClearAndSelect)
        with signal_waiter(self._data_connection.file_system_watcher.file_added) as waiter:
            self._data_connection.copy_to_project()
            waiter.wait()
        self.assertEqual(self._ref_model.rowCount(ref_root_index), 0)
        self.assertTrue(Path(self._toolbox.project().items_dir, "dc", "a.txt").exists())
        self.assertEqual(self._data_connection.data_model.rowCount(), 1)
        index = self._data_connection.data_model.index(0, 0)
        self.assertEqual(index.data(), "a.txt")
        self.assertEqual(
            index.data(Qt.ItemDataRole.UserRole + 1), os.path.join(self._toolbox.project().items_dir, "dc", "a.txt")
        )

    def test_create_data_file(self):
        with mock.patch("spine_items.data_connection.data_connection.QInputDialog") as mock_input_dialog:
            mock_input_dialog.getText.return_value = ["data.csv"]
            with signal_waiter(self._data_connection.file_system_watcher.file_added) as waiter:
                self._data_connection.make_new_file()
                waiter.wait()
        model = self._data_connection.data_model
        self.assertEqual(model.rowCount(), 1)
        index = model.index(0, 0)
        self.assertEqual(index.data(), "data.csv")
        self.assertEqual(
            index.data(Qt.ItemDataRole.UserRole + 1), os.path.join(self._toolbox.project().items_dir, "dc", "data.csv")
        )

    def test_deleting_data_file_removes_it_from_dc(self):
        file_a = Path(self._data_connection.data_dir) / "data.dat"
        with signal_waiter(self._data_connection.file_system_watcher.file_added) as waiter:
            file_a.touch()
            waiter.wait()
        model = self._data_connection.data_model
        self.assertEqual(model.rowCount(), 1)
        index = model.index(0, 0)
        self.assertEqual(index.data(), "data.dat")
        self.assertEqual(index.data(Qt.ItemDataRole.UserRole + 1), str(file_a))
        with signal_waiter(self._data_connection.file_system_watcher.file_removed) as waiter:
            file_a.unlink()
            waiter.wait()
        self.assertEqual(model.rowCount(), 0)

    def test_renaming_data_file_renames_it_in_dc_as_well(self):
        file_a = Path(self._data_connection.data_dir) / "data.dat"
        with signal_waiter(self._data_connection.file_system_watcher.file_added) as waiter:
            file_a.touch()
            waiter.wait()
        model = self._data_connection.data_model
        self.assertEqual(model.rowCount(), 1)
        index = model.index(0, 0)
        self.assertEqual(index.data(), "data.dat")
        self.assertEqual(index.data(Qt.ItemDataRole.UserRole + 1), str(file_a))
        renamed = file_a.parent / "sata.txt"
        with signal_waiter(self._data_connection.file_system_watcher.file_renamed) as waiter:
            file_a.rename(renamed)
            waiter.wait()
        self.assertEqual(model.rowCount(), 1)
        index = model.index(0, 0)
        self.assertEqual(index.data(), "sata.txt")
        self.assertEqual(index.data(Qt.ItemDataRole.UserRole + 1), str(renamed))

    def test_item_dict(self):
        """Tests Item dictionary creation."""
        d = self._data_connection.item_dict()
        a = ["type", "description", "x", "y", "file_references", "db_references", "db_credentials"]
        for k in a:
            self.assertTrue(k in d, f"Key '{k}' not in dict {d}")

    def test_deserialization_with_remote_db_reference(self):
        with (
            mock.patch("spine_items.data_connection.data_connection.UrlSelectorDialog.exec") as url_selector_exec,
            mock.patch(
                "spine_items.data_connection.data_connection.UrlSelectorDialog.url_dict"
            ) as url_selector_url_dict,
        ):
            # Add nothing
            url_selector_exec.return_value = QDialog.DialogCode.Accepted
            url_selector_url_dict.return_value = {
                "dialect": "mysql",
                "host": "post.com",
                "port": 3306,
                "database": "db",
                "username": "randy",
                "password": "creamfraiche",
            }
            self._data_connection.show_add_db_reference_dialog()
        item_dict = self._data_connection.item_dict()
        self.assertEqual(len(item_dict["db_references"]), 1)
        self.assertNotIn("username", item_dict["db_references"][0])
        self.assertNotIn("password", item_dict["db_references"][0])
        deserialized = DataConnection.from_dict("deserialized", item_dict, self._toolbox, self._toolbox.project())
        self.assertTrue(deserialized.has_db_references())
        self.assertEqual(
            list(deserialized.db_reference_iter()),
            [
                {
                    "dialect": "mysql",
                    "host": "post.com",
                    "port": 3306,
                    "database": "db",
                    "username": "randy",
                    "password": "creamfraiche",
                }
            ],
        )

    def test_deserialization_with_sqlite_db_reference_in_project_directory(self):
        db_path = Path(self._temp_dir.name, "db.sqlite")
        create_new_spine_database("sqlite:///" + str(db_path))
        with (
            mock.patch("spine_items.data_connection.data_connection.UrlSelectorDialog.exec") as url_selector_exec,
            mock.patch(
                "spine_items.data_connection.data_connection.UrlSelectorDialog.url_dict"
            ) as url_selector_url_dict,
        ):
            # Add nothing
            url_selector_exec.return_value = QDialog.DialogCode.Accepted
            url_selector_url_dict.return_value = {
                "dialect": "sqlite",
                "host": None,
                "port": None,
                "database": str(db_path),
                "username": None,
                "password": None,
            }
            self._data_connection.show_add_db_reference_dialog()
        item_dict = self._data_connection.item_dict()
        self.assertEqual(len(item_dict["db_references"]), 1)
        self.assertNotIn("username", item_dict["db_references"][0])
        self.assertNotIn("password", item_dict["db_references"][0])
        deserialized = DataConnection.from_dict("deserialized", item_dict, self._toolbox, self._toolbox.project())
        self.assertTrue(deserialized.has_db_references())
        self.assertEqual(
            list(deserialized.db_reference_iter()),
            [
                {
                    "dialect": "sqlite",
                    "host": None,
                    "port": None,
                    "database": str(db_path),
                    "username": None,
                    "password": None,
                }
            ],
        )

    def test_sqlite_db_reference_is_marked_missing_when_db_file_is_renamed(self):
        db_path = Path(self._temp_dir.name, "db.sqlite")
        create_new_spine_database("sqlite:///" + str(db_path))
        with (
            mock.patch("spine_items.data_connection.data_connection.UrlSelectorDialog.exec") as url_selector_exec,
            mock.patch(
                "spine_items.data_connection.data_connection.UrlSelectorDialog.url_dict"
            ) as url_selector_url_dict,
        ):
            # Add nothing
            url_selector_exec.return_value = QDialog.DialogCode.Accepted
            url_selector_url_dict.return_value = {
                "dialect": "sqlite",
                "host": None,
                "port": None,
                "database": str(db_path),
                "username": None,
                "password": None,
            }
            self._data_connection.show_add_db_reference_dialog()
        while self._data_connection._database_validator.is_busy():
            QApplication.processEvents()
        self.assertEqual(
            list(self._data_connection.db_reference_iter()),
            [
                {
                    "dialect": "sqlite",
                    "host": None,
                    "port": None,
                    "database": str(db_path),
                    "username": None,
                    "password": None,
                }
            ],
        )
        gc.collect()
        with signal_waiter(self._data_connection.file_system_watcher.file_renamed) as waiter:
            db_path.rename(db_path.parent / "renamed.sqlite")
            waiter.wait()
        self.assertTrue(self._data_connection._db_ref_root.child(0, 0).data(_Role.MISSING))

    def test_refreshing_missing_sqlite_reference_resurrects_it(self):
        db_path = Path(self._temp_dir.name, "db.sqlite")
        create_new_spine_database("sqlite:///" + str(db_path))
        with (
            mock.patch("spine_items.data_connection.data_connection.UrlSelectorDialog.exec") as url_selector_exec,
            mock.patch(
                "spine_items.data_connection.data_connection.UrlSelectorDialog.url_dict"
            ) as url_selector_url_dict,
        ):
            # Add nothing
            url_selector_exec.return_value = QDialog.DialogCode.Accepted
            url_selector_url_dict.return_value = {
                "dialect": "sqlite",
                "host": None,
                "port": None,
                "database": str(db_path),
                "username": None,
                "password": None,
            }
            self._data_connection.show_add_db_reference_dialog()
        while self._data_connection._database_validator.is_busy():
            QApplication.processEvents()
        self.assertEqual(
            list(self._data_connection.db_reference_iter()),
            [
                {
                    "dialect": "sqlite",
                    "host": None,
                    "port": None,
                    "database": str(db_path),
                    "username": None,
                    "password": None,
                }
            ],
        )
        gc.collect()
        with signal_waiter(self._data_connection.file_system_watcher.file_renamed) as waiter:
            renamed_path = db_path.rename(db_path.parent / "renamed.sqlite")
            waiter.wait()
        self.assertTrue(self._data_connection._db_ref_root.child(0, 0).data(_Role.MISSING))
        with signal_waiter(self._data_connection.file_system_watcher.file_renamed) as waiter:
            renamed_path.rename(db_path)
            waiter.wait()
        self.assertTrue(self._data_connection._db_ref_root.child(0, 0).data(_Role.MISSING))
        self._data_connection.restore_selections()
        self._data_connection._connect_signals()
        db_ref_root_index = self._ref_model.index(1, 0)
        ref_index = self._ref_model.index(0, 0, db_ref_root_index)
        self._data_connection._properties_ui.treeView_dc_references.selectionModel().select(
            ref_index, QItemSelectionModel.Select
        )
        self._data_connection.refresh_references()
        while self._data_connection._database_validator.is_busy():
            QApplication.processEvents()
        self.assertFalse(self._data_connection._db_ref_root.child(0, 0).data(_Role.MISSING))

    def test_broken_sqlite_url_marks_the_reference_missing(self):
        db_path = Path(self._temp_dir.name, "db.sqlite")
        with (
            mock.patch("spine_items.data_connection.data_connection.UrlSelectorDialog.exec") as url_selector_exec,
            mock.patch(
                "spine_items.data_connection.data_connection.UrlSelectorDialog.url_dict"
            ) as url_selector_url_dict,
        ):
            # Add nothing
            url_selector_exec.return_value = QDialog.DialogCode.Accepted
            url_selector_url_dict.return_value = {
                "dialect": "sqlite",
                "host": None,
                "port": None,
                "database": str(db_path),
                "username": None,
                "password": None,
            }
            self._data_connection.show_add_db_reference_dialog()
        while self._data_connection._database_validator.is_busy():
            QApplication.processEvents()
        self.assertEqual(
            list(self._data_connection.db_reference_iter()),
            [
                {
                    "dialect": "sqlite",
                    "host": None,
                    "port": None,
                    "database": str(db_path),
                    "username": None,
                    "password": None,
                }
            ],
        )
        self.assertTrue(self._data_connection._db_ref_root.child(0, 0).data(_Role.MISSING))

    def test_notify_destination(self):
        self._data_connection.logger.msg = MagicMock()
        self._data_connection.logger.msg_warning = MagicMock()
        source_item = NonCallableMagicMock()
        source_item.name = "source name"
        source_item.item_type = MagicMock(return_value="Importer")
        self._data_connection.notify_destination(source_item)
        self._data_connection.logger.msg.emit.assert_called_with("Link established")
        source_item.item_type = MagicMock(return_value="Data Store")
        self._data_connection.notify_destination(source_item)
        self._data_connection.logger.msg.emit.assert_called_with("Link established")
        source_item.item_type = MagicMock(return_value="Tool")
        self._data_connection.notify_destination(source_item)
        self._data_connection.logger.msg.emit.assert_called_with(
            "Link established. Tool <b>source name</b> output files"
            " will be passed as references to item <b>DC</b> after execution."
        )
        source_item.item_type = MagicMock(return_value="View")
        self._data_connection.notify_destination(source_item)
        self._data_connection.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>View</b> and"
            " a <b>Data Connection</b> has not been implemented yet."
        )

    def test_rename(self):
        """Tests renaming a Data Connection."""
        self._data_connection.activate()
        expected_name = "ABC"
        expected_short_name = "abc"
        expected_data_dir = os.path.join(self._toolbox.project().items_dir, expected_short_name)
        self._data_connection.rename(expected_name, "")
        # Check name
        self.assertEqual(expected_name, self._data_connection.name)
        self.assertEqual(expected_name, self._data_connection.get_icon().name())
        # Check data_dir
        self.assertEqual(expected_data_dir, self._data_connection.data_dir)
        # Check that file_system_watcher has one path (new data_dir)
        watched_dirs = self._data_connection.file_system_watcher.directories()
        self.assertEqual(1, len(watched_dirs))
        self.assertEqual(self._data_connection.data_dir, watched_dirs[0])


class TestDataConnectionWithInitialDataFile(unittest.TestCase):
    def setUp(self):
        """Set up."""
        self.toolbox = create_mock_toolbox()
        factory = DataConnectionFactory()
        item_dict = {"type": "Data Connection", "description": "", "references": [], "x": 0, "y": 0}
        self._temp_dir = TemporaryDirectory()
        self.project = create_mock_project(self._temp_dir.name)
        self.toolbox.project.return_value = self.project
        self._item_dir = Path(self.project.items_dir, "dc")
        self._item_dir.mkdir(parents=True)
        self._data_file_path = self._item_dir / "data.csv"
        self._data_file_path.touch()
        self.data_connection = factory.make_item("DC", item_dict, self.toolbox, self.project)
        self._properties_tab = mock_finish_project_item_construction(factory, self.data_connection, self.toolbox)

    def tearDown(self):
        self.data_connection.tear_down()
        self._temp_dir.cleanup()

    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_data_file_in_list(self):
        model = self.data_connection.data_model
        self.assertEqual(model.rowCount(), 1)
        index = model.index(0, 0)
        self.assertEqual(index.data(), "data.csv")
        self.assertEqual(index.data(Qt.ItemDataRole.UserRole + 1), str(self._data_file_path))

    def test_remove_data_file(self):
        self.data_connection.restore_selections()
        index = self.data_connection.data_model.index(0, 0)
        self._properties_tab.ui.treeView_dc_data.selectionModel().select(index, QItemSelectionModel.ClearAndSelect)
        with mock.patch("spine_items.data_connection.data_connection.QMessageBox") as mock_message_box:
            mock_message_box.exec.return_value = QMessageBox.StandardButton.Ok
            with signal_waiter(self.data_connection.file_system_watcher.file_removed) as waiter:
                self.data_connection.remove_files()
                waiter.wait()
        model = self.data_connection.data_model
        self.assertEqual(model.rowCount(), 0)

    def test_rename_data_file(self):
        with signal_waiter(self.data_connection.file_system_watcher.file_renamed) as waiter:
            self._data_file_path.rename(self._item_dir / "renamed.dat")
            waiter.wait()
        model = self.data_connection.data_model
        self.assertEqual(model.rowCount(), 1)
        index = model.index(0, 0)
        self.assertEqual(index.data(), "renamed.dat")
        self.assertEqual(index.data(Qt.ItemDataRole.UserRole + 1), str(self._item_dir / "renamed.dat"))


class TestDataConnectionWithInvalidFileReference(unittest.TestCase):
    def setUp(self):
        """Set up."""
        self._temp_dir = TemporaryDirectory()
        self._non_existent_path = os.path.join(self._temp_dir.name, "file.dat")
        self.toolbox = create_mock_toolbox()
        factory = DataConnectionFactory()
        item_dict = {
            "type": "Data Connection",
            "description": "",
            "file_references": [self._non_existent_path],
            "x": 0,
            "y": 0,
        }
        self.project = create_mock_project(self._temp_dir.name)
        self.toolbox.project.return_value = self.project
        self.data_connection = factory.make_item("DC", item_dict, self.toolbox, self.project)
        self.project.get_item.return_value = self.data_connection
        self._properties_tab = mock_finish_project_item_construction(factory, self.data_connection, self.toolbox)
        self.ref_model = self.data_connection.reference_model

    def tearDown(self):
        self.data_connection.tear_down()
        self._temp_dir.cleanup()

    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_add_file_references(self):
        temp_dir = Path(self._temp_dir.name, "references")
        temp_dir.mkdir()
        with mock.patch("spine_items.data_connection.data_connection.QFileDialog.getOpenFileNames") as mock_filenames:
            a = Path(temp_dir, "a.txt")
            a.touch()
            mock_filenames.return_value = ([str(a)], "*.*")
            self.data_connection.show_add_file_references_dialog()
            self.assertEqual(2, len(self.data_connection.file_references))
            self.assertEqual(2, self.ref_model.rowCount(self.ref_model.index(0, 0)))
            self.assertEqual(self._non_existent_path, self.ref_model.index(0, 0, self.ref_model.index(0, 0)).data())
            self.assertEqual(str(a), self.ref_model.index(1, 0, self.ref_model.index(0, 0)).data())

    def test_refresh_file_reference(self):
        root_index = self.ref_model.index(0, 0)
        reference_index = self.ref_model.index(0, 0, root_index)
        self.assertEqual(reference_index.data(Qt.ItemDataRole.ToolTipRole), "The file is missing.")
        Path(self._non_existent_path).touch()
        self.data_connection.restore_selections()
        self._properties_tab.ui.treeView_dc_references.selectionModel().select(
            reference_index, QItemSelectionModel.ClearAndSelect
        )
        self.data_connection.refresh_references()
        self.assertIsNone(reference_index.data(Qt.ItemDataRole.ToolTipRole), "The file is missing.")
        self.project.notify_resource_changes_to_successors.assert_called_once_with(self.data_connection)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

################################################################################
## Form generated from reading UI file 'exporter_properties.ui'
##
## Created by: Qt User Interface Compiler version 6.7.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QFrame,
    QHBoxLayout, QLabel, QScrollArea, QSizePolicy,
    QSpacerItem, QToolButton, QVBoxLayout, QWidget)
from spine_items import resources_icons_rc

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(272, 202)
        Form.setStyleSheet(u"QScrollArea { background: transparent; }\n"
"QScrollArea > QWidget > QWidget { background: transparent; }")
        self.verticalLayout_3 = QVBoxLayout(Form)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.scrollArea = QScrollArea(Form)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setFrameShape(QFrame.Shape.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 272, 202))
        self.verticalLayout = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_9 = QHBoxLayout()
        self.horizontalLayout_9.setSpacing(4)
        self.horizontalLayout_9.setObjectName(u"horizontalLayout_9")
        self.specification_label = QLabel(self.scrollAreaWidgetContents)
        self.specification_label.setObjectName(u"specification_label")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.specification_label.sizePolicy().hasHeightForWidth())
        self.specification_label.setSizePolicy(sizePolicy)
        self.specification_label.setMaximumSize(QSize(16777215, 16777215))

        self.horizontalLayout_9.addWidget(self.specification_label)

        self.specification_combo_box = QComboBox(self.scrollAreaWidgetContents)
        self.specification_combo_box.setObjectName(u"specification_combo_box")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.specification_combo_box.sizePolicy().hasHeightForWidth())
        self.specification_combo_box.setSizePolicy(sizePolicy1)

        self.horizontalLayout_9.addWidget(self.specification_combo_box)

        self.specification_button = QToolButton(self.scrollAreaWidgetContents)
        self.specification_button.setObjectName(u"specification_button")
        icon = QIcon()
        icon.addFile(u":/icons/wrench.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.specification_button.setIcon(icon)
        self.specification_button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)

        self.horizontalLayout_9.addWidget(self.specification_button)


        self.verticalLayout.addLayout(self.horizontalLayout_9)

        self.message_label = QLabel(self.scrollAreaWidgetContents)
        self.message_label.setObjectName(u"message_label")
        self.message_label.setWordWrap(True)

        self.verticalLayout.addWidget(self.message_label)

        self.outputs_list_layout = QVBoxLayout()
        self.outputs_list_layout.setSpacing(0)
        self.outputs_list_layout.setObjectName(u"outputs_list_layout")

        self.verticalLayout.addLayout(self.outputs_list_layout)

        self.verticalSpacer = QSpacerItem(20, 71, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.frame = QFrame(self.scrollAreaWidgetContents)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.StyledPanel)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_2 = QVBoxLayout(self.frame)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.output_time_stamps_check_box = QCheckBox(self.frame)
        self.output_time_stamps_check_box.setObjectName(u"output_time_stamps_check_box")

        self.verticalLayout_2.addWidget(self.output_time_stamps_check_box)

        self.cancel_on_error_check_box = QCheckBox(self.frame)
        self.cancel_on_error_check_box.setObjectName(u"cancel_on_error_check_box")
        self.cancel_on_error_check_box.setChecked(True)

        self.verticalLayout_2.addWidget(self.cancel_on_error_check_box)


        self.verticalLayout.addWidget(self.frame)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.verticalLayout_3.addWidget(self.scrollArea)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.specification_label.setText(QCoreApplication.translate("Form", u"Specification:", None))
#if QT_CONFIG(tooltip)
        self.specification_combo_box.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Tool specification for this Tool</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.specification_button.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Edit specification.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.message_label.setText("")
#if QT_CONFIG(tooltip)
        self.output_time_stamps_check_box.setToolTip(QCoreApplication.translate("Form", u"Checking this will add time stamps to output directory names.", None))
#endif // QT_CONFIG(tooltip)
        self.output_time_stamps_check_box.setText(QCoreApplication.translate("Form", u"Time stamp output directories", None))
        self.cancel_on_error_check_box.setText(QCoreApplication.translate("Form", u"Cancel export on error", None))
    # retranslateUi




######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains base classes for items that write to db."""
from spine_engine.project_item.executable_item_base import ExecutableItemBase
from spinedb_api.spine_db_client import SpineDBClient


class DBWriterExecutableItemBase(ExecutableItemBase):
    """Base class for items that might write to a Spine DB."""

    def exclude_execution(self, forward_resources, backward_resources, lock):
        """Perform the checkout on output databases so concurrent items can proceed."""
        to_resources = [r for r in backward_resources if r.type_ == "database"]
        for resource in to_resources:
            resource.quick_db_checkout()

    def update(self, forward_resources, backward_resources):
        if not super().update(forward_resources, backward_resources):
            return False
        to_resources = [r for r in backward_resources if r.type_ == "database"]
        for resource in to_resources:
            with resource.open() as server_url:
                SpineDBClient.from_server_url(server_url).cancel_db_checkout()
        return True



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains classes to manage entity class renaming."""
from PySide6.QtCore import QObject, QSortFilterProxyModel, Qt, Slot
from ..commands import InsertRow, RemoveRow
from ..mvcmodels.class_renames_table_model import ClassRenamesTableModel
from ..settings import EntityClassRenamingSettings
from .copy_paste import copy_table_data, paste_table_data


class ClassRename(QObject):
    _MIME_TYPE = "application/spine-dtclassrename"

    def __init__(self, ui, undo_stack, settings, parent):
        """
        Args:
            ui (Ui_Form): specification editor's UI
            undo_stack (QUndoStack): undo stack
            settings (EntityClassRenamingSettings, optional): initial settings
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._ui = ui
        self._undo_stack = undo_stack
        name_map = settings.name_map if isinstance(settings, EntityClassRenamingSettings) else {}
        self._rename_table_model = ClassRenamesTableModel(undo_stack, name_map)
        self._sorted_rename_table_model = QSortFilterProxyModel(self)
        self._sorted_rename_table_model.setSortCaseSensitivity(Qt.CaseInsensitive)
        self._sorted_rename_table_model.setSourceModel(self._rename_table_model)
        self._ui.class_rename_table_view.setModel(self._sorted_rename_table_model)
        self._ui.class_rename_table_view.addAction(self._ui.remove_class_rename_action)
        self._ui.add_class_button.clicked.connect(self._add_class)
        self._ui.remove_class_rename_action.triggered.connect(self._remove_class)
        self._ui.remove_class_button.clicked.connect(self._ui.remove_class_rename_action.trigger)
        self._ui.class_rename_table_view.addAction(self._ui.copy_class_rename_data_action)
        self._ui.copy_class_rename_data_action.triggered.connect(self._copy_table_data)
        self._ui.class_rename_table_view.addAction(self._ui.paste_class_rename_data_action)
        self._ui.paste_class_rename_data_action.triggered.connect(self._paste_table_data)

    @Slot(bool)
    def _add_class(self, checked):
        """Pushes an add class command to undo stack.

        Args:
            checked (bool): unused
        """
        row = self._rename_table_model.rowCount()
        self._undo_stack.push(InsertRow("add class", self._rename_table_model, row, ["class", ""]))

    @Slot(bool)
    def _remove_class(self, checked):
        """Pushes a remove class command to undo stack.

        Args:
            checked (bool) unused
        """
        indexes = self._ui.class_rename_table_view.selectionModel().selectedIndexes()
        if not indexes:
            return
        rows = set(self._sorted_rename_table_model.mapToSource(i).row() for i in indexes)
        if len(rows) == 1:
            self._undo_stack.push(RemoveRow("remove class", self._rename_table_model, next(iter(rows))))
        else:
            self._undo_stack.beginMacro("remove classes")
            for row in reversed(sorted(rows)):
                self._undo_stack.push(RemoveRow("", self._rename_table_model, row))
            self._undo_stack.endMacro()

    def load_data(self, url):
        """
        Loads entity class names from given URL.

        Args:
            url (str): database URL
        """
        self._ui.available_classes_tree_widget.load_data(url)

    @Slot(bool)
    def _copy_table_data(self, checked):
        """Copies data to clipboard.

        Args:
            checked (bool): unused
        """
        copy_table_data(self._ui.class_rename_table_view, self._MIME_TYPE)

    @Slot(bool)
    def _paste_table_data(self, checked):
        """Pastes data from clipboard.

        Args:
            checked (bool): unused
        """
        paste_table_data(self._ui.class_rename_table_view, self._MIME_TYPE, self._undo_stack)

    def settings(self):
        """
        Returns filter's settings.

        Returns:
            FilterSettings: settings
        """
        return EntityClassRenamingSettings(self._rename_table_model.renaming_settings())

    def show(self):
        """Shows docks."""
        self._ui.load_database_dock.show()
        self._ui.possible_classes_dock.show()
        self._ui.class_rename_dock.show()

    def tear_down(self):
        """Hides docks and releases resources."""
        self._ui.load_database_dock.hide()
        self._ui.possible_classes_dock.hide()
        self._ui.class_rename_dock.hide()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains the :class:`ClassRenamesTableModel` class."""
from enum import IntEnum, unique
import pickle
from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt
from ..commands import InsertRow, SetData
from ..widgets.drop_target_table import DROP_MIME_TYPE


@unique
class ClassTableColumn(IntEnum):
    CLASS = 0
    NEW_NAME = 1


@unique
class RenamesRoles(IntEnum):
    SILENT_EDIT = Qt.ItemDataRole.UserRole + 1


class ClassRenamesTableModel(QAbstractTableModel):
    """A table model for entity class renaming tables."""

    GET_DATA_ROLES = (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.DisplayRole)
    SET_DATA_ROLES = (RenamesRoles.SILENT_EDIT, RenamesRoles.SILENT_EDIT)

    def __init__(self, undo_stack, renaming):
        """
        Args:
            undo_stack (QUndoStack)
            renaming (dict): renaming settings
        """
        super().__init__()
        self._undo_stack = undo_stack
        self._renames = [[original, renamed] for original, renamed in renaming.items()]

    def columnCount(self, parent=QModelIndex()):
        """
        Returns:
            int: number of columns
        """
        return 2

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        """Returns table data for given role."""
        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            return self._renames[index.row()][index.column()]
        return None

    def dropMimeData(self, data, action, row, column, parent):
        if row < 0:
            row = self.rowCount()
        classes = pickle.loads(data.data(DROP_MIME_TYPE))
        rows = [[klass, ""] for klass in classes]
        if len(rows) == 1:
            self._undo_stack.push(InsertRow("add class", self, row, rows[0]))
        else:
            self._undo_stack.beginMacro("add classes")
            for i, row_data in enumerate(rows):
                self._undo_stack.push(InsertRow("", self, row + i, row_data))
            self._undo_stack.endMacro()
        return True

    def mimeTypes(self):
        return [DROP_MIME_TYPE]

    def flags(self, index):
        return super().flags(index) | Qt.ItemIsDropEnabled | Qt.ItemIsEditable

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        """Returns header data."""
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            return ("Original", "Renamed")[section]
        return None

    def insertRows(self, row, count, parent=QModelIndex()):
        rows = count * [["", ""]]
        self.beginInsertRows(parent, row, row + count - 1)
        self._renames = self._renames[:row] + rows + self._renames[row:]
        self.endInsertRows()
        return True

    def removeRows(self, row, count, parent=QModelIndex()):
        self.beginRemoveRows(parent, row, row + count - 1)
        self._renames = self._renames[:row] + self._renames[row + count :]
        self.endRemoveRows()
        return True

    def renaming_settings(self):
        return {row[ClassTableColumn.CLASS]: row[ClassTableColumn.NEW_NAME] for row in self._renames}

    def rowCount(self, parent=QModelIndex()):
        """
        Returns:
            int: number of rows
        """
        return len(self._renames)

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if not index.isValid():
            return False
        column = index.column()
        if role == Qt.ItemDataRole.EditRole:
            old_value = self._renames[index.row()][column]
            if value == old_value:
                return False
            message = "change class name" if column == ClassTableColumn.CLASS else "change new class name"
            self._undo_stack.push(SetData(message, index, value, old_value, RenamesRoles.SILENT_EDIT))
            return True
        if role == RenamesRoles.SILENT_EDIT:
            self._renames[index.row()][index.column()] = value
            self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])
            return True
        return False



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################
import unittest
from unittest import mock
from PySide6.QtWidgets import QApplication
from spine_items.widgets import UrlSelectorWidget
from tests.mock_helpers import parent_widget


class TestUrlSelectorWidget(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_first_dialect_is_selected_in_combo_box(self):
        with parent_widget() as parent:
            widget = UrlSelectorWidget(parent)
            callback = mock.MagicMock()
            logger = mock.MagicMock()
            widget.setup(["dialect 1", "dialect 2"], callback, False, logger)
            widget.set_url({})
            self.assertEqual(widget._ui.comboBox_dialect.currentIndex(), 0)


if __name__ == "__main__":
    unittest.main()
